package StreamSimulator.Buffer;

/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */
import java.util.HashMap;
import java.util.Vector;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.*;

import java.io.FileWriter;
import java.io.IOException;

import StreamSimulator.Stream.*;
import StreamSimulator.StreamJoin.StreamJoin;

public class AgingBuffer extends JoinBuffer implements StreamBuffer {

  private double creditRatio;
  public boolean seperateBuffer;
  public AgingBuffer(int size, double creditRatio, boolean seperateBuffer, int num_streams) {
    super(size,num_streams);
    this.creditRatio = creditRatio;
    this.seperateBuffer= seperateBuffer;
  }

  public synchronized boolean insert(BufferTuple tuple)
  {
    boolean res = false;
      if (this.currentSize < this.bufferSize) 
      {
        res = true;
        // try to find if there exists a tuple in the buffer with the same attribute
        int index = this.lookupCredit(tuple);
        // if so, copy its credit and insert the tuple into the buffer
        if (index != -1) {
          // if it were CONSTANT_TOTAL_CREDIT method, compensate the adjustment 
          if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT) {
	          double value = ((BufferTuple)buffer.elementAt(index)).credit;
	          buffer.insertElementAt(tuple, index+1);
	          compensate(0-(value/buffer.size()));
		  }
	      else    
	          buffer.insertElementAt(tuple, index + 1);
          
        }
        // else copy the credit of the tuple at a certain percentile to this new tuple
        else if (this.currentSize > 0)
        {
          if (this.seperateBuffer)
          {
          	double tmp = StreamJoin.AGING_INITIAL_VALUE*(double)this.currentSize;
            index = (int)tmp;
          }
          else
          {
              int[] streamcountHere = this.countStreamSize();
              int threshHold =  (int)(StreamJoin.AGING_INITIAL_VALUE*streamcountHere[tuple.stream_ID]);
              /*int count =0;
              int size = this.buffer.size();
              for (int i=size-1;i>=0;i--)
              {
                BufferTuple tpl = (BufferTuple)buffer.elementAt(i);
                if (tpl.stream_ID == tuple.stream_ID)
                   count++;
                if (count > threshHold)
                {
                  index =i;
                  break;
                }
              }*/
              index = threshHold;
          }
          if (index==-1)
            index=0;

		  if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT) {
		  	  double c = ((BufferTuple)buffer.elementAt(index)).credit;
	          tuple.credit = c;
	          compensate(0-c);
	      }
	      else
	          tuple.credit = ((BufferTuple)buffer.elementAt(index)).credit;

          buffer.insertElementAt(tuple, index);
        }
        else {
		  if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT)
	          tuple.credit = 0.0;
	      tuple.credit=0.0;
          buffer.add(tuple);
        }
      }
      currentSize++;

    return res;
  }

  public synchronized long[] join(BufferTuple tuple, Topology top)
  {
    if (top.seperate_buffer == false)
    {
     	if (tuple.stream_ID==0)
     		return join_UnionBufferTest(tuple,1);
     	else if (tuple.stream_ID==1)
     		return join_UnionBufferTest(tuple,0);
     	else if (tuple.stream_ID==2)
     		return join_UnionBufferTest(tuple,3);
     	else if (tuple.stream_ID==3)
    		return join_UnionBufferTest(tuple,2);
     	//return join_UnionBuffer(tuple);
    } 
    return join_SeperateBuffer(tuple,top);
        /*if (Math.abs(tpl.latitude-tuple.latitude)<100 && Math.abs(tpl.longtitude
          -tuple.longtitude)<100)
        {
            if (Math.abs(tpl.value - tuple.value)<10)
            {

              tpl.credit +=  Math.pow(size - index, creditRatio);
              results ++;
            }
          }*/
          /*
          if (tpl.value == tuple.value || tpl.objID == tuple.objID)
          {
             if (tpl.value == tuple.value && tpl.objID == tuple.objID)
             {
                tpl.credit +=  2*Math.pow(size - index, creditRatio);
                results+=2;
              }
              else
              {
                 tpl.credit +=  Math.pow(size - index, creditRatio);
                 results++;
              }
          }*/
  }

  public synchronized long[] join_SeperateBuffer(BufferTuple tuple, Topology top)
  {
  
	FileWriter fw = null;
	try {
		fw = new FileWriter("joinOuptut_tmp",true);
	} catch (IOException e) {
		e.printStackTrace();
	}

  
    long[] results = new long[StreamJoin.NUM_ATTRIBUTE];
    int[][] count = new int[this.num_streams][StreamJoin.NUM_ATTRIBUTE];
    for (int i = 0; i < this.num_streams; i++)
      for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        count[i][j] = 0;
    for (int i = 0; i < StreamJoin.NUM_ATTRIBUTE; i++)
      results[i] = 1;

    for (int i=0;i<top.numBuffer;i++)
    {
      if (i == tuple.stream_ID)
        continue;
      int size = top.buffer[i].buffer.size();
      for (int k = 0; k < size; k++)
      {
        if (k >= top.buffer[i].buffer.size())
          break;
        try
        {
          BufferTuple tpl = (BufferTuple) top.buffer[i].buffer.elementAt(k);
          for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++) {
            if (tpl.value[j] == tuple.value[j]) {

            	try {
            		fw.write(tpl.value[j] + "\r\n");
            	} catch (IOException e) {
            		e.printStackTrace();
            	}            

              count[i][j]++;
            }
          }
        }catch(Exception e)
        {
          e.printStackTrace();
          break;
        }
      }
    }
    for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
    {
      for (int i = 0; i < this.num_streams; i++)
      {
        if (i != tuple.stream_ID)
          results[j] *= count[i][j];
      }
    }

    for (int i = 0; i < top.numBuffer; i++)
    {
      if (i==tuple.stream_ID)
        continue;
      int size = top.buffer[i].buffer.size();
      for (int k=0; k<size;k++)
      {
        if (k >= top.buffer[i].buffer.size())
          break;
        try
        {
          BufferTuple tpl = (BufferTuple) top.buffer[i].buffer.elementAt(k);
          for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++) {
            if (tpl.value[j] == tuple.value[j]) {
              if (count[i][j] != 0) {
                int joinTimes = (int) results[j] / count[i][j];
               	if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_BY_OFFSET)
                  tpl.credit = tpl.credit +
                     Math.pow(size - k, creditRatio);
                else if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_AND_MINUS_ONE)
	                tpl.credit+=1;
	            else if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT)
	            	tpl.credit+=1;
              }
            }
          }
        }catch(Exception e)
        {
          e.printStackTrace();
          break;
        }
      }
    }
    
/*
	// DEBUG    
    for(int j = 0; j < count.length; j++) {
    	for(int i = 0; i < count[j].length; i++) {
    		System.out.print(count[j][i] + " ");
    	}
    	System.out.println("");
    }
*/
    
    try {
	    fw.close();
	} catch (IOException e) {
		e.printStackTrace();
	}
    
    
    return results;
  }
  
   public synchronized long[] join_UnionBufferTest(BufferTuple tuple,int dd)
  {
    int size = buffer.size();
    long[] results = new long[StreamJoin.NUM_ATTRIBUTE];
    int[][] count = new int[this.num_streams][StreamJoin.NUM_ATTRIBUTE];
    for (int i = 0; i < this.num_streams; i++)
      for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        count[i][j] = 0;
    for (int i = 0; i < StreamJoin.NUM_ATTRIBUTE; i++)
      results[i] = 1;
    for (int i = 0; i < size; i++)
    {
      BufferTuple tpl = (BufferTuple) buffer.elementAt(i);
      if (tpl.stream_ID == dd)
      {
        for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        {
          if (tpl.value[j] == tuple.value[j])
            count[tpl.stream_ID][j]++;
        }
      }
    }
    for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
    {
      for (int i = 0; i < this.num_streams; i++)
      {
        if (i == dd)
          results[j] = count[i][j];
      }
    }
    int[] streamcountHere = this.countStreamSize();
    int[] streamSeenCount = new int[this.num_streams];
    for (int i=0;i<this.num_streams;i++)
      streamSeenCount[i]=0;

    for (int i = 0; i < size; i++)
    {
      BufferTuple tpl = (BufferTuple) buffer.elementAt(i);
      streamSeenCount[tpl.stream_ID]++;
      if (tpl.stream_ID == dd)
      {
        for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        {
          if (tpl.value[j] == tuple.value[j])
          {
            int joinTimes = (int) results[j] / count[tpl.stream_ID][j];
            if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_BY_OFFSET)
                  tpl.credit = tpl.credit +
                      Math.pow(size - i, creditRatio)/streamcountHere[tpl.stream_ID];
            else if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_AND_MINUS_ONE)
	                tpl.credit+=1;
	        else if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT)
	            	tpl.credit+=1;
          }
        }
      }
    }
    return results;
  }
  
  


  

  public synchronized long[] join_UnionBuffer(BufferTuple tuple)
  {
    int size = buffer.size();
    long[] results = new long[StreamJoin.NUM_ATTRIBUTE];
    int[][] count = new int[this.num_streams][StreamJoin.NUM_ATTRIBUTE];
    for (int i = 0; i < this.num_streams; i++)
      for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        count[i][j] = 0;
    for (int i = 0; i < StreamJoin.NUM_ATTRIBUTE; i++)
      results[i] = 1;
    for (int i = 0; i < size; i++)
    {
      BufferTuple tpl = (BufferTuple) buffer.elementAt(i);
      if (tpl.stream_ID != tuple.stream_ID)
      {
        for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        {
          if (tpl.value[j] == tuple.value[j])
            count[tpl.stream_ID][j]++;
        }
      }
    }
    for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
    {
      for (int i = 0; i < this.num_streams; i++)
      {
        if (i != tuple.stream_ID)
          results[j] *= count[i][j];
      }
    }
    int[] streamcountHere = this.countStreamSize();
    int[] streamSeenCount = new int[this.num_streams];
    for (int i=0;i<this.num_streams;i++)
      streamSeenCount[i]=0;

    for (int i = 0; i < size; i++)
    {
      BufferTuple tpl = (BufferTuple) buffer.elementAt(i);
      streamSeenCount[tpl.stream_ID]++;
      if (tpl.stream_ID != tuple.stream_ID)
      {
        for (int j = 0; j < StreamJoin.NUM_ATTRIBUTE; j++)
        {
          if (tpl.value[j] == tuple.value[j])
          {
            int joinTimes = (int) results[j] / count[tpl.stream_ID][j];
            if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_BY_OFFSET)
                  tpl.credit = tpl.credit +
                      Math.pow(size - i, creditRatio)/streamcountHere[tpl.stream_ID];
            else if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_AND_MINUS_ONE)
	                tpl.credit+=1;
	        else if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT)
	            	tpl.credit+=1;
          }
        }
      }
    }
    return results;
  }


  public synchronized void drop(Topology top)
  {
    BufferTuple tuple=null;
    /*int i=1;
    tuple=(BufferTuple)buffer.elementAt(currentSize-i);
    while (tuple.stat==BufferTuple.NEW_ARRIVAL && i<=currentSize)
    {
      i++;
      tuple=(BufferTuple)buffer.elementAt(currentSize-i);
    }
    if (i<=currentSize)
    {
      buffer.remove(tuple);
    }
    else
    {
      BufferTuple tpl = (BufferTuple)buffer.remove(currentSize - 1);
    }*/
    
    if (top.seperate_buffer == true)
    {
	    if(StreamJoin.AGING_METHODS == StreamJoin.CONSTANT_TOTAL_CREDIT) {
	    
	    	double value = ((BufferTuple)buffer.elementAt(currentSize-1)).credit;
		    buffer.remove(currentSize-1);
		    compensate(value/buffer.size());
			currentSize--;
			
	    } else {      	
	    	buffer.remove(currentSize-1);
		    //if (this.currentSize>0)
	        currentSize--;
	    	   
	    }
	}
	else
	{
/*
		double rateRatio = top.avgArrivalRate[0]/top.avgArrivalRate[1];
		int[] streamcountHere = this.countStreamSize();

		double countRatio = streamcountHere[0]/streamcountHere[1];
		int toDrop = 0;
		if (countRatio > rateRatio)
			toDrop = 0;
		else
			toDrop = 1;
	    int i=1;
    	tuple=(BufferTuple)buffer.elementAt(currentSize-i);
	    while (tuple.stream_ID!=toDrop)
    	{
	      i++;
    	  tuple=(BufferTuple)buffer.elementAt(currentSize-i);
    	}
	    buffer.remove(currentSize-i);
	    currentSize--;
*/	    

		buffer.remove(currentSize - 1);
		currentSize--;

	}	
    
  }

  public synchronized void aging(long t)
  {
    int size = buffer.size();
   	double tCredit =0.0;
	double drops = 0.0;
	if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_AND_MINUS_ONE)
		return;
	if (StreamJoin.AGING_METHODS == StreamJoin.CONSTANT_TOTAL_CREDIT)
	{ 	
		tCredit = totalCredit();

	    if(Math.abs(tCredit) < 0.1E-3)
    	  	tCredit = 0.0; 
	/*
		if(tCredit > 10000000) {
		    for (int i=0;i<size;i++)
		    {
	    	    BufferTuple tp = (BufferTuple)buffer.elementAt(i);
				System.out.print(tp.credit + " ");
		    }
    	
	    }
		*/	
		
		if (this.currentSize!=0)
			drops = tCredit/(double)this.currentSize;
//		System.out.println("total credit:"+tCredit+" "+drops+" "+this.buffer.size());
	}

    for (int i=0;i<size;i++)
    {
      BufferTuple tpl = (BufferTuple)buffer.elementAt(i);
      if (!this.seperateBuffer)
      {
          if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_BY_OFFSET)
          {
	          int[] streamcountHere = this.countStreamSize();
	    	  double[] drop = new double[this.num_streams];
		      for (int j=0;j<this.num_streams;j++)
      				drop[j] = (double)streamcountHere[j]/(double)this.buffer.size();
		      tpl.credit = (tpl.credit-drop[tpl.stream_ID]);                   
	      }
          else if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_AND_MINUS_ONE)
	         tpl.credit--;
	      else if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT)
	      {
	      	 tpl.credit-=drops;	
	      }
      }
      else
      {
          if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_BY_OFFSET)
	        tpl.credit--;
          else if (StreamJoin.AGING_METHODS==StreamJoin.PLUS_AND_MINUS_ONE)
	        tpl.credit--;
	      else if (StreamJoin.AGING_METHODS==StreamJoin.CONSTANT_TOTAL_CREDIT)
	      {
	      	 tpl.credit-=drops;	
	      }
      }
    }
  }

  public synchronized void orderBuffer()
  {

    Collections.sort(buffer);

  }

  public synchronized double totalCredit()
  {
    int size = buffer.size();
    double totalCredit=0.0;
    for (int i=0;i<size;i++)
    {
      BufferTuple tp = (BufferTuple)buffer.elementAt(i);
      if(Math.abs(tp.credit) < 0.1E-3)
      	tp.credit = 0.0; 
      totalCredit+=tp.credit;
    }
    return totalCredit;
  }

  public synchronized void organizeBuffer(int size)
  {
   /*
    if (size>=currentSize)
       this.bufferSize = size;
    else
    {
      int toDrop = currentSize - size;
      BufferTuple tuple = null;
      for (int i=1;i<=toDrop;i++)
        tuple = (BufferTuple)buffer.remove(currentSize-i);
      this.currentSize = size;
      this.bufferSize = size;
    }
    */
  }
}
