package StreamSimulator.Buffer;

/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */

import java.util.HashMap;
import java.util.Vector;
import java.util.Iterator;
import java.util.Enumeration;
import StreamSimulator.Stream.*;
import StreamSimulator.StreamJoin.*;
import StreamSimulator.MinSkewPacket.MinSkew2D.*;
import StreamSimulator.MinSkewPacket.MinSkew4D.*;

public class ProbabilitySheddingBuffer extends JoinBuffer implements StreamBuffer {

  public ProbabilitySheddingBuffer(int size, int num_streams)
  {
    super(size,num_streams);
  }

  public synchronized void drop(Topology top)
  {
    int selected=0;
    if (this.use_histogram == false)
      selected = drop_byProduct(top.statistics);
    else
      selected = drop_byHistogram(top.histograms);
    //System.out.println(selected);
    buffer.remove(selected);
    this.currentSize--;
  }

  public synchronized int drop_byProduct(HashMap statistics) {
    double min = 0;
    int selected = 0;
    for (int i = 0; i < this.currentSize; i++)
    {
      BufferTuple tuple = (BufferTuple) buffer.elementAt(i);
      //if (tuple.stat == BufferTuple.NEW_ARRIVAL)
      //continue;
      for (int j = 0; j < this.num_streams; j++)
      {
         if (j == tuple.stream_ID)
            continue;
        double combinedPro = 0.0;
        double[] probability = new double[StreamJoin.NUM_ATTRIBUTE];
        String keyTotal = j+":total";
        Long total = (Long)statistics.get(keyTotal);
        for (int k=0; k< StreamJoin.NUM_ATTRIBUTE;k++)
        {
          String key = j + ":" +k+":"+tuple.value[k];
          Object obj = statistics.get(key);
          if (obj==null)
            probability[k]=0;
          else
          {
              Long count = (Long)obj;
              probability[k] = count.doubleValue()/total.doubleValue();
          }
        }
        for (int z=0;z<StreamJoin.NUM_ATTRIBUTE;z++)
            combinedPro+=probability[z];
        if (StreamJoin.NUM_ATTRIBUTE == 2)
        {
          combinedPro+=2*probability[0]*probability[1];
        }
        if (StreamJoin.NUM_ATTRIBUTE == 3)
        {
          combinedPro+=2*(probability[0]*probability[1]+probability[0]*probability[2]
                          +probability[1]*probability[2]);
          combinedPro+=3*(probability[0]*probability[1]*probability[2]);
        }
        if (StreamJoin.NUM_ATTRIBUTE == 4)
        {
          combinedPro+=2*(probability[0]*probability[1]+probability[0]*probability[2]
                          +probability[0]*probability[3]+probability[1]*probability[2]
                          +probability[1]*probability[3]+probability[2]*probability[3]);
          combinedPro+=3*(probability[0]*probability[1]*probability[2]+
                          probability[0]*probability[1]*probability[3]+
                          probability[1]*probability[2]*probability[3]);
          combinedPro+=4*(probability[0]*probability[1]*probability[2]*probability[3]);
        }
        if (StreamJoin.NUM_ATTRIBUTE == 5)
        {
          combinedPro+=2*(probability[0]*probability[1]+probability[0]*probability[2]
                          +probability[0]*probability[3]+probability[0]*probability[4]+
                          probability[1]*probability[2]+probability[1]*probability[3]+probability[1]*probability[4]
                          +probability[2]*probability[3]+probability[2]*probability[4]+probability[3]*probability[4]);
          combinedPro+=3*(probability[0]*probability[1]*probability[2]+
                          probability[0]*probability[1]*probability[3]+
                          probability[0]*probability[1]*probability[4]+
                          probability[1]*probability[2]*probability[3]+
                          probability[1]*probability[2]*probability[4]+
                          probability[2]*probability[3]*probability[4]);
          combinedPro+=4*(probability[0]*probability[1]*probability[2]*probability[3]+probability[0]*probability[1]*probability[2]*probability[4]
          				  +probability[1]*probability[2]*probability[3]*probability[4]);
          combinedPro+=5*(probability[0]*probability[1]*probability[2]*probability[3]*probability[4]);
        }

        /*
        double[][] tmp = new double[probability.length][probability.length];
        for (int z=0;z<probability.length;z++)
        {
          for (int q=0;q<probability.length;q++)
            tmp[z][q] = probability[q];
        }
        for (int k=1;k<=probability.length;k++)
          combinedPro+=i*streamJoin.combinationGen(-1,1.0,0,k, tmp[k-1]);
        */
       if (combinedPro == 0.0)
         return i;
        if (i == 0 && j == 0)
        {
          min = combinedPro;
          selected = 0;
        }
        if (combinedPro < min)
        {
          min = combinedPro;
          selected = i;
        }
      }
    }
    //System.out.println(selected);
    return selected;
  }

  public synchronized int drop_byHistogram(Vector[] histograms)
  {
    long min = 0;
    int selected = 0;
    for (int i = 0; i < this.currentSize; i++)
    {
      BufferTuple tuple = (BufferTuple) buffer.elementAt(i);
      for (int j = 0; j < this.num_streams; j++)
      {
        if (j == tuple.stream_ID)
          continue;
        long count = 0;
        if (StreamJoin.NUM_ATTRIBUTE == 2)
          count = MinSkew2D.checkHistogram(tuple,histograms[j]);
        else if (StreamJoin.NUM_ATTRIBUTE == 4)
          count = MinSkew4D.checkHistogram(tuple,histograms[j]);
        if (count == -1)
          return i;
        if (i==0 && j==0)
        {
          min = count;
          selected = i;
        }
        if (count < min)
        {
          min = count;
          selected = i;
        }
      }
    }
    //System.out.println(selected);
    return selected;
  }

}
