package StreamSimulator.Buffer;

import java.io.File;

import StreamSimulator.Graph.KOpt;
import StreamSimulator.StreamJoin.StreamJoin;
import StreamSimulator.Stream.Topology;
import StreamSimulator.StreamGenerator.Generator;

public class StanfordBuffer extends JoinBuffer implements StreamBuffer {

	// enumerate method types
	private final int STAGGER = 0;	
	private final int FIFO = 1;
	
	// average k optimal in time unit of the streams for every period
	private static int[] k1;
	private static int[] k2;
	private int[] k;
	
	// average k optimal in time unit times arrival rate of the stream for every period
	private static int[] real_k1;
	private static int[] real_k2;
	private int[] real_k;
	
	// starting time for a new period for every period
	private static int[] start_time1;
	private static int[] start_time2;
	private int[] start_time;
	
	// the arrival rate of both stream for every period
	private static double[] arrival_rate1;
	private static double[] arrival_rate2;
	private double[] arrival_rate;

	// total number of periods
	private int periods;

	// track the current tuple
	private int current_tuple;

	// method type depending on window size, k optimal and tuples arrival rate
	private int type = 0;

	// signal if the members have already been initialized
	private static boolean init = false;

	// Constructor	
	public StanfordBuffer(int size, int num_streams, int which) {
		super(size, num_streams);
			
		// if the static variables are not initialized, do it now
		if(!init) {

			init = true;

			// Calculate k optimal value
			File file1 = new File("data0");
			File file2 = new File("data1");

			KOpt kopt = null;
			if(StreamJoin.STREAM_TYPE == Generator.GISMO) {
				kopt = new KOpt(file1, file2, 
					StreamJoin.WINDOW_SIZE, StreamJoin.UNIT, 1, 2);
			} else if(StreamJoin.STREAM_TYPE == Generator.STOCK) {
				kopt = new KOpt(file1, file2, 
					StreamJoin.WINDOW_SIZE, StreamJoin.UNIT, 0, 1);
			} else if(StreamJoin.STREAM_TYPE == Generator.WEATHER) {
				kopt = new KOpt(file1, file2,
					StreamJoin.WINDOW_SIZE, StreamJoin.UNIT, -1, -1);
			} else {
				System.out.println("Stanford buffer only support Gismo, Weather and Stock data at the moment.");
				return;
			}
		
			// get the k optimal value for each tuple for each stream	
			long[] optCurve1 = kopt.getOptCurve(1);
			long[] optCurve2 = kopt.getOptCurve(2);
		
			// number of periods in the whole life time of the stream
			periods = StreamJoin.STANFORD_PERIOD;
		
			// initialize all the arrays
			int accum1;
			int accum2;
			int index = 0;
			int average1 = 0;
			int average2 = 0;
			double average_rate1 = 0;
			double average_rate2 = 0;				
			int count1 = 0;
			int count2 = 0;				
			k1 = new int[optCurve1.length];
			k2 = new int[optCurve2.length];
			real_k1 = new int[optCurve1.length];
			real_k2 = new int[optCurve2.length];
			arrival_rate1 = new double[optCurve1.length];
			arrival_rate2 = new double[optCurve2.length];
		
			for(int i = 0; i<Math.max(optCurve1.length,optCurve2.length); i++) {
				average1 = 0;
				average2 = 0;

				average_rate1 = 0;
				average_rate2 = 0;
					
				count1 = 0;
				count2 = 0;
					
				accum1 = 0;
				accum2 = 0;
			
				// average the kopt value in this period
				for(int j = index; j<index+periods; j++) {
					if(j < optCurve1.length) {
						count1++;
						average1 += optCurve1[j];
					}
					if(j < optCurve2.length) {
						count2++;
						average2 += optCurve2[j];
					}
				}

				// calculate the arrival rate of this period
				int prev_index;
				if(index == 0)
					prev_index = 0;
				else prev_index = index-1;
				
				if(index < optCurve1.length && index+count1-1 > 0) {
					if(kopt.stream1[index+count1-1][1] - kopt.stream1[prev_index][1] <= 0)
						average_rate1 = 0.0;
					else
						average_rate1 = (periods * StreamJoin.UNIT) / (double)
										( kopt.stream1[index+count1-1][1] - kopt.stream1[prev_index][1]);
				}

				if(index < optCurve2.length && index+count2-1 > 0) {
					if(kopt.stream2[index+count2-1][1] - kopt.stream2[prev_index][1] <= 0)
						average_rate2 = 0.0;
					else
						average_rate2 = (periods * StreamJoin.UNIT) / (double)
										(kopt.stream2[index+count2-1][1] - kopt.stream2[prev_index][1]);
				}

				if(count1 == 0)
					average1 = 0;
				else average1 /= count1;
				
				if(count2 == 0)
					average2 = 0;
				else average2 /= count2;
				
				// computer the real k - tuple arrival rate multiply with the kopt of that tuple
				for(int j = index; j<index+periods; j++) {
					if(j < optCurve1.length) {
						arrival_rate1[j] = average_rate1;
						k1[j] = average1;
						real_k1[j] = (int)average_rate1 * average1;
					}		
					if(j < optCurve2.length) {
						arrival_rate2[j] = average_rate2;
						k2[j] = average2;
						real_k2[j] = (int)average_rate2 * average2;
					}
				}

				// update the index for the next round
				index += periods;
			}
		} // end if init is false

		// set the current period to begin at 0
		current_tuple = 0;

		// assign the values to the specific stream
		if(which == 1) {
			k = k1;
			real_k = real_k1;
			arrival_rate = arrival_rate1;
		} else if(which == 2) {
			k = k2;
			real_k = real_k2;
			arrival_rate = arrival_rate2;
		} else {
			System.out.println(which + ": Wrong stream index!");
		}
	
		// if k optimal is greater than the buffer size
		if(real_k[current_tuple] > bufferSize) {
			// use staggering method
			type = STAGGER;
		System.out.println("Stanford method: STAGGERING");
		} else {
			// else use FIFO
			type = FIFO;
			System.out.println("Stanford method: FIFO");
		}

	} // end Constructor
	


	// drop - StreamBuffer interface function
	//  select a tuple to drop from the buffer
	public synchronized void drop(Topology top) {
		// if FIFO, simply removed the earliest element
		if(type == FIFO) {
			buffer.remove(0);
			currentSize--;
		// else, stagger
		} else {
			double minCredit = 1.0;
			int minCreditIndex = -1;
			
			// check for a tuple in the buffer that has resided more than k unit time
			for(int i = 0; i<buffer.size(); i++) {
				BufferTuple tuple = (BufferTuple)buffer.elementAt(i);
				// the tuple that has stayed for the longest time beyond k should have the smallest credit
				if(tuple.credit <= 0) {
					if(tuple.credit < minCredit) {
						minCredit = tuple.credit;
						minCreditIndex = i;
					}
				}
			}		

			if(minCreditIndex != -1) {			
				buffer.remove(minCreditIndex);
				currentSize--;
			}
		}

	}	

	// insert - StreamBuffer interface function
	//  add a tuple into the buffer	
	// update k optimal value and algorithm types (FIFO/STAGGER)
	public synchronized boolean insert(BufferTuple tuple) {
		if (currentSize < bufferSize) {
			tuple.credit = k[current_tuple];

			// if current time exceeds the current period, update the k and algo
			if(real_k[current_tuple] > bufferSize) {
				// use staggering method
				type = STAGGER;
				System.out.println("Stanford method: STAGGERING");
			} else {
				// else use FIFO
				type = FIFO;
				System.out.println("Stanford method: FIFO");
			}

			System.out.println("Tuple " + current_tuple + ": " + k[current_tuple] + " * " + 
								arrival_rate[current_tuple] + " = " + real_k[current_tuple]);

			buffer.add(tuple);
			currentSize++;
			
			current_tuple++;
			return true;
		}
		
		// full buffer, skip this tuple and return false
		current_tuple++;		
		return false;
	}

	public synchronized void aging(long t) {
		for(int i = 0; i<buffer.size(); i++) {
			BufferTuple tuple = (BufferTuple)buffer.elementAt(i);
			tuple.credit -= 1.0;
		}		
	}

}
