package StreamSimulator.Generator;

import java.util.Vector;
import java.util.Random;
import java.lang.Math;

public class ZipfGenerator {
  private double a;
  private int ord;
  private int c;
  private int seed;
  
  private double locality;

  // ord - number of distinct values
  // c - muliplication constant in zipf
  //   - this is also use to adjust the size of the sample.
  // a - the power constant in zipf
  // seed - a seed for random number generation
  public ZipfGenerator(int ord, int c, double a, int seed) {
  	this.ord = ord;
  	this.c = c;
    this.a = a;
    this.seed = seed;
    this.locality = 1.0/ord;
  }

  public Vector nextData() {
    Vector output = new Vector();
    Random rand = new Random(seed);
  	int[] distr = new int[ord];
  	int[] locality_set = new int[ord];
  	int skewedTotal;
  	int key;
  	int index;
    int total = 0;
  	for(int i = 0; i<ord; i++) {
  	    distr[i] = (int)(c*(1/(Math.pow(i+1,a))));
  		total += distr[i];
  	}
  	
  	for(int i = 0; i<ord; i++) 
  		locality_set[i] = 0;
  	
  	for(int i = 0; i<c; i++) {
  		skewedTotal = getSkewedTotal(distr,locality_set,ord);
  		key = rand.nextInt(skewedTotal);

  		for(index = 0; key>0; index++)
  			key-=distr[index]*(1+(locality_set[index]*locality));
  		output.add(new Integer(index));
  		//update locality set
  		for(int j = 0; j<ord; j++) {
  			if(locality_set[j]>0)
  				locality_set[j]--;
  		}
  		locality_set[index]=ord-1;
  	}
  	return output;
  }
  
  public int getSkewedTotal(int[] distr, int[] locality_set, int ord) {
  	double skewedTotal = 0;
  	for(int i = 0; i<ord; i++)
  		skewedTotal += distr[i]*(1+(locality_set[i]*locality));

  	return (int)skewedTotal;
  }

/*  public Vector nextData()
  {
    float key;
    double index;
    int i;
  	Vector output = new Vector();
  	Random rand = new Random(seed);
  	
  	double locality = 0.5/ord;
  	
  	int[] distr = new int[ord];
  	int total = 0;
  	
  	for(i = 0; i<ord; i++) {
  		distr[i]=c*(1/(Math.pow(i+1,a)));
  		total+=distr[i];
  	}
  	
  	while(total>0) {
  		key = rand.nextFloat();
  		index = key*total;

  		for(i = 0; index>0; i++)
  			index-=distr[i]*(1+(locality*(i-1)));

  		output.add(new Integer(i));
  		distr[i]--;
  		total--;
  	}

  	return output;
  }*/
}

