package StreamSimulator.Graph;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileNotFoundException;

import java.util.Calendar;
import java.util.TimeZone;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.StringTokenizer;

import StreamSimulator.StreamJoin.StreamJoin;
import StreamSimulator.StreamGenerator.WeatherDataGenerator;

import java.text.SimpleDateFormat;


public class KOpt {
	
	// arrays of stream data {objectID, time}
	public long[][] stream1;
	public long[][] stream2;
	
	// sXprobe for each tuple in each stream, i.e. Pi(j) value
//	private long[][] s1probe;
//	private long[][] s2probe;
    private int probeCount = 1;
	
	// rate, i.e. C(k)/k for each 0<k<=W for each tuple in each stream
	private double[][] rate1;
	private double[][] rate2;
	
	// k optimal value for each tuple in each stream
	private long[] opt1;
	private long[] opt2;

	// k optimal value for each stream
	private long kopt1;
	private long kopt2;

	// arrive rate for each stream
	public double arrivalRate1;
	public double arrivalRate2;
	
	private final String DATE_FORMAT = "HH-mm-ss-SSS";
	private String timeStr = "";

	public KOpt() {}


	// Constructor
	// read file and parse them into arrays
	// calculate k optimal and the intermediate results for each tuple in each streams	
	// unit: time unit
	// obj: column index of the obj in the file
	// time: column index of time in the file
	public KOpt(File file1, File file2, int W, int unit, int obj, int time) {
	
		Calendar cal = Calendar.getInstance(TimeZone.getDefault());
		SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
		sdf.setTimeZone(TimeZone.getDefault());
		timeStr = sdf.format(cal.getTime());

		String stamp =  StreamJoin.WINDOW_SIZE + "-" + StreamJoin.BUFFERS[0] + "-" + 
						StreamJoin.BUFFERS[1] + "-" + StreamJoin.TIME_RANGE + "-" +
						StreamJoin.UNIT + "-" + StreamJoin.STREAM_TYPE + "-" + timeStr;
							
		// read the file and format it into stream arrays {objectID, time}
		if(obj != -1) { // not weather data
			stream1 = parse(file1, obj, time);
			stream2 = parse(file2, obj, time);
			try {
				saveToFile("saveStream1"+stamp+".txt", stream1);
				saveToFile("saveStream2"+stamp+".txt", stream2);
			} catch (IOException e) {
				e.printStackTrace();
			}
			
		} else {
			stream1 = parse(file1);
			stream2 = parse(file2);	
			try {		
				saveToFile("saveStream1"+stamp+".txt", stream1);
				saveToFile("saveStream2"+stamp+".txt", stream2);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		// compute the arrive rates
		arrivalRate1 = (stream1.length*unit)/(double)(stream1[stream1.length-1][1]-stream1[0][1]);
		arrivalRate2 = (stream2.length*unit)/(double)(stream2[stream2.length-1][1]-stream2[0][1]);

		System.out.println("finished parsing");

		// probe each tuple to the other stream to get sXprobe value for every time unit		
		probe(stream1, stream2, W, unit);		

		System.out.println("finished probing 1");
		
		probe(stream2, stream1, W, unit);
		
		System.out.println("finished probing 2");
		
		// calculate the rate at each time instance for each tuple in each stream
		int r1 = rateCk(1);

		System.out.println("finished calculating rate 1");
		
		int r2 = rateCk(2);
		
		System.out.println("finished calculating rate 2");
		
		// get the optimal k value (maximum rate) for each tuple in each stream
		opt1 = optimalK(r1, 1);
		opt2 = optimalK(r2, 2);

		// get the average optimal k value for each stream		
		kopt1 = mean(opt1);
		kopt2 = mean(opt2);
		
		System.out.println("finished calculating k optimal");
				
		/* DEBUG */
		try {
				
//			saveToFile("saveS1Probe"+stamp+".txt", s1probe);
//			saveToFile("saveS2Probe"+stamp+".txt", s2probe);
//			saveToFile("saveRate1"+stamp+".txt", rate1);
//			saveToFile("saveRate2"+stamp+".txt", rate2);
			saveToFile("saveOpt1"+stamp+".txt", opt1);
			saveToFile("saveOpt2"+stamp+".txt", opt2);
			
			long[] temp = new long[2];
			temp[0] = kopt1;
			temp[1] = kopt2;
			saveToFile("saveKOpt"+stamp+".txt", temp);

			System.out.println( "kopt1: " + kopt1 + " kopt2: " + kopt2 );
			System.out.println( "arrival rate 1: " + arrivalRate1 + "\n" +
								"arrival rate 2: " + arrivalRate2);
								
		} catch (IOException e) {
			e.printStackTrace();
		}

	}
	
	// save matrix to a file
	private void saveToFile(String fname, double[][] data) throws IOException {
		File save = new File(fname);
		if(save.exists()) {
			save.delete();
		}
		FileWriter fw = new FileWriter(save);
		for(int i = 0; i<data.length; i++) {
			if(data[i] == null) break;
			for(int j = 0; j<data[i].length; j++) {
				fw.write(data[i][j] + " ");
			}
			fw.write("\r\n");
		}
		fw.close();
	}
	
	// save array to a file
	private void saveToFile(String fname, long[] data) throws IOException  {
		File save = new File(fname);
		if(save.exists()) {
			save.delete();
		}
		FileWriter fw = new FileWriter(save);
		for(int i = 0; i<data.length; i++) {
			fw.write(data[i] + "\r\n");
		}
		fw.close();
	}

	// save matrix to a file
	private void saveToFile(String fname, long[][] data) throws IOException {
		File save = new File(fname);
		if(save.exists()) {
			save.delete();
		}
		FileWriter fw = new FileWriter(save);
		for(int i = 0; i<data.length; i++) {
			if(data[i] == null) break;
			for(int j = 0; j<data[i].length; j++) {
				fw.write(data[i][j] + " ");
			}
			fw.write("\r\n");
		}
		fw.close();
	}

	// parse a file into stream tuple array	
	private long[][] parse(File file, int obj, int time) {
		BufferedReader in = null;
		ArrayList arr = new ArrayList();
		
		try {
	    	in = new BufferedReader(new FileReader(file));
		} catch (FileNotFoundException e) { e.printStackTrace(); }
	
	    try {
	    
			while(true) {
				String s = in.readLine();			
				if(s == null || s.compareTo("")==0) break;
				
				StringTokenizer tok = new StringTokenizer(s, "\t\n\r\f, ");
				long[] l = new long[tok.countTokens()];
				long[] tuple = new long[2];

				for(int i = 0; i<l.length; i++) {
					l[i] = Long.parseLong(tok.nextToken());
				}
			
				tuple[0] = l[obj];
				tuple[1] = l[time];
				arr.add(tuple);
			}
		} catch (IOException e) { e.printStackTrace(); }
					
		long[][] stream = new long[arr.size()][2];
		for(int i = 0; i<arr.size(); i++) {
			long[] tuple = (long[])arr.get(i);
			stream[i][0] = tuple[0];
			stream[i][1] = tuple[1];
		}	

		return stream;	
	}

	// parse a file into stream tuple array	
	private long[][] parse(File file) {
		BufferedReader in = null;
		ArrayList arr = new ArrayList();
		
		try {
	    	in = new BufferedReader(new FileReader(file));
		} catch (FileNotFoundException e) { e.printStackTrace(); }
	
	    try {
	    	    
			while(true) {
				String s = in.readLine();			
				long[] tuple = new long[2];
				if(s == null || s.compareTo("")==0) break;

				StringTokenizer tok = new StringTokenizer(s, "\t\n\r\f, ");
		        tuple[1] = WeatherDataGenerator.formatDate(tok.nextToken(), tok.nextToken());
		        tok.nextToken();
    		    tok.nextToken();
        		tok.nextToken();
		        tuple[0] = (int)Double.parseDouble(tok.nextToken());

				arr.add(tuple);
			}
		} catch (IOException e) { e.printStackTrace(); }
					
		long[][] stream = new long[arr.size()][2];
		for(int i = 0; i<arr.size(); i++) {
			long[] tuple = (long[])arr.get(i);
			stream[i][0] = tuple[0];
			stream[i][1] = tuple[1];
		}	

		return stream;	
	}
	
	// calculate sXprobe value for every "eligible" tuple (tuples that arrive
	// after time W from the beginning of the stream and before time W from the
	// end of the stream) at every time unit with W window size
	private void probe(long[][] prober, long[][] probed, int W, int unit) {
		int count = 0;
		long maxTime = prober[prober.length-1][1];
		int last_visited = 0;
		
		FileWriter fw = null;
		try {
			fw = new FileWriter("probe" + "-" + timeStr + probeCount);
		} catch (IOException e) {
			e.printStackTrace();
		}
		probeCount++;
			
		for(int i = 0; i<prober.length; i++) {
			ArrayList arr = new ArrayList();
			if(prober[i][1] > maxTime)
				break;

			// update last_visited to skip over historic data
			for(int k = last_visited; k<probed.length; k++) {
				if(probed[k][1] < prober[i][1]) {
					last_visited = k;
				} else break;
			}
			
			int last_visited2 = last_visited;
		
			try {
				// "eligible" tuples here
				for(int j = 0; j<W; j++) {
					if(j*unit+prober[i][1] <= maxTime) {
						fw.write(countJoin(last_visited, prober[i][0], probed, 
									j*unit+prober[i][1], j*unit+prober[i][1]+unit) + " ");
					} 
					else break;
				}
				last_visited = last_visited2;			
				fw.write("\r\n");
			} catch (IOException e) {
				e.printStackTrace();
			}
/*		
			output[count] = new long[arr.size()];
			for(int j = 0; j<arr.size(); j++)
				output[count][j] = ((Long)arr.get(j)).longValue();
			count++;
*/		
		}		
		
		try {
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return;
	
	}
	
	// count the number of join produced by "key" on "table" from "start" time to "end" time
	private long countJoin(int last_visited, long key, long[][] table, long start, long end) {
		long count = 0;
		for(int i = last_visited; i<table.length; i++) {
			if(table[i][1] < start) {
				last_visited = (int)table[i][1];
				continue;
			}
			if(table[i][1] >= end)
				break;
			
			if(table[i][0] == key)
				count++;
		}
	
		return count;
	}
	
	// calculate the rate - C(k)/k for each 0<k<=W, for each "eligible" tuple
	private int rateCk(int probe) {
	
		int count = 0;
		int sum = 0;

		BufferedReader in = null;
		FileWriter fw = null;

		try {
		
	    	in = new BufferedReader(new FileReader("probe" + "-" + timeStr + probe));
	    	fw = new FileWriter("rate" + "-" + timeStr + probe);
	    	
		} catch (Exception e) { e.printStackTrace(); }
	
	    try {
	    	    
			while(true) {
				String s = in.readLine();			
				if(s == null || s.compareTo("")==0) break;
				count++;
				
				StringTokenizer tok = new StringTokenizer(s, "\t\n\r\f, ");
				long[] tuple = new long[tok.countTokens()];
				sum = 0;
				for(int i = 0; i<tuple.length; i++) {
					tuple[i] = Long.parseLong(tok.nextToken());
					sum += tuple[i];
					fw.write(sum/(double)(i+1) + " ");
				}
				fw.write("\r\n");
				
			}
		} catch (IOException e) { e.printStackTrace(); }
				
		try {
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return count;	
	}

	// calculate the k optimal values (index of the maximum rate) for each eligible tuple 
	private long[] optimalK(int count, int which) {
		long[] output = new long[count];
		BufferedReader in = null;

		try {
	    	in = new BufferedReader(new FileReader("rate" + "-" + timeStr + which));
		} catch (FileNotFoundException e) { e.printStackTrace(); }

		try {
			for(int i = 0; i<count; i++) {
				String s = in.readLine();
				if(s == null || s.compareTo("")==0) break;
				StringTokenizer tok = new StringTokenizer(s, "\t\n\r\f, ");
				double rate[] = new double[tok.countTokens()];
				for(int j = 0; j<rate.length; j++) {
					rate[j] = Double.parseDouble(tok.nextToken());
				}
				output[i] = maxIndex(rate);

				// discredit tuples with very low rate
				if(output[i] != 0 && rate[(int)output[i]-1]<StreamJoin.STANFORD_THRESHOLD)
					output[i] = 0;

			}
		} catch (IOException e) {
			e.printStackTrace();
		}
			
		return output;
	}
	
	// get the array of k optimal values for every tuples in a stream
	public long[] getOptCurve(int which) {
		if(which == 1)
			return opt1;
		else if(which == 2)
			return opt2;
		else
			System.out.println(which + ": Wrong value for getOptCurve, input is neither 1 or 2");
		return null;
	}
	
	// get the average k optimal value for a specified stream
	public long getOpt(int which) {
		if(which == 1)
			return kopt1;
		else if(which == 2)
			return kopt2;
		else 
			System.out.println("Wrong value for getOpt, input is neither 1 or 2");
		return 0;
	}
	
	// return the age curve of an "eligible" tuple specified by index
	/*
	public long[] getCurve(int which, int index) {
		if(which == 1)
			return s2probe[index];
		else if(which == 2)
			return s1probe[index];
		else 
			System.out.println(which + ": Wrong value for getCurve, input is neither 1 or 2");
		return null;
	}
	*/

	// calculate the max of a long array
	private long maxIndex(double[] num) {
		double m = num[0];
		long mindex = 1;

		for(int i = 1; i<num.length; i++) {
			if(num[i] > m) {
				m = num[i];
				mindex = i+1;
			}
		}

		if(mindex == 1 && num[0] == 0.0)
			return 0;
		return mindex;
	}
	
	// calculate the mean of a long array
	private long mean(long[] num) {
		double sum = 0;
		
		for(int i = 0; i<num.length; i++) {
			sum += (double)num[i]/(double)num.length;
		}
		
		return (long)sum;
	}
		
} // KOpt