package StreamSimulator.MinSkewPacket.MinSkew4D;

/**
 * <p>Title: MinSkew Algorithms</p>
 * <p>Description: For StreamJoin Project</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Computer Science Dept. Boston University</p>
 * @author lifeifei
 * @version 1.0
 */
 
import StreamSimulator.StreamJoin.StreamJoin;
import StreamSimulator.MinSkewPacket.MinSkew4D.*;
import StreamSimulator.MinSkewPacket.*;
import StreamSimulator.Buffer.*;
import java.util.*;
import java.lang.Math;

public class MinSkew4D {

  public int DIM;
  public MinSkew4D() {
    this.DIM = MinBucket.DIM;
  }

  public Vector createHistogram(Vector data, int Xi, int NumbOfBins)
  {
        int i,j;
        int dataLen = data.size();
        int i1, i2, i3, i4;
        int   low, high;
        long  numofpoints;
        double[] mond = new double[DIM];
        int d1, d2, d3, d4;
        MinBucket b1 = new MinBucket();
        int k;

        long RealPoints;
        int  NumOfBuckets;
        double  TotalCells;
        double GlobalAverage, GlobalVariance;
        int  t1, t2, t3, t4;
        int  NumberOfBuckets;
        int  BuckToSplit;
        int  SplitDimension;
        int  BestSplitPoint;
        double  BestVarDecrease;
        double[] AvgDim= new double[DIM];
        double[] VarDim= new double[DIM];
        double SmallestVar;
        int  SmallestDim;
        int  splitp;
        double  leftvalue;
        int  nleft;
        double  rightvalue;
        double avgleft;
        int  totaln;
        int nright;
        double avgright;
        double NewSplitVariance;
        double VarDecrease;
        int  NewBValue;
        int  NewCellNumb;
        double AvgNewB;
        int  OldCellNum;
        double AvgOldB;
        double  NewBuckVariance;
        double  OldBuckVariance;
        double  NewSplitVarianceRight;
        double  NewSplitVarianceLeft;

        int[][][][] CellArray= new int[Xi + 1][Xi + 1][Xi + 1][Xi + 1];
        MinBucket[] BucketArray = new MinBucket[NumbOfBins];
        Interval[] DimLimits= new Interval[DIM];

        //added by feifei, initialize all object array
        for (i=0;i<NumbOfBins;i++)
          BucketArray[i] = new MinBucket();
        for (i=0;i<DIM;i++)
          DimLimits[i] = new Interval();

        for (i = 0; i < DIM; i++)
        {
                DimLimits[i].h = -1000000.0;
                DimLimits[i].l = 1000000.0;
        }

        // find dimension limits.
        for (k=0;k<dataLen;k++)
        {
          Point4D p = (Point4D)data.elementAt(k);
          for (i = 0; i < DIM; i++) {
            DimLimits[i].h = Math.max(DimLimits[i].h, p.data[i]);
            DimLimits[i].l = Math.min(DimLimits[i].l, p.data[i]);
          }
        }

        numofpoints = dataLen;
        RealPoints = numofpoints;

        mond[0] = (double)(DimLimits[0].h - DimLimits[0].l)/(double)Xi;
        mond[1] = (double)(DimLimits[1].h - DimLimits[1].l)/(double)Xi;
        mond[2] = (double)(DimLimits[2].h - DimLimits[2].l)/(double)Xi;
        mond[3] = (double)(DimLimits[3].h - DimLimits[3].l)/(double)Xi;

        for (i1=0; i1 < Xi; i1++)
        {
                for (i2=0; i2 < Xi; i2++)
                {
                        for (i3=0; i3 < Xi; i3++)
                        {
                                for (i4=0; i4 < Xi; i4++)
                                {
                                        CellArray[i1][i2][i3][i4] = 0;
                                }
                        }
                }
        }

        /* run through all data points */
        for (k=0;k<dataLen;k++)
        {
          Point4D p = (Point4D)data.elementAt(k);
          d1 = Math.min((int)((p.data[0] - DimLimits[0].l) / mond[0]), Xi-1);
          d2 = Math.min((int)((p.data[1] - DimLimits[1].l) / mond[1]), Xi-1);
          d3 = Math.min((int)((p.data[2] - DimLimits[2].l) / mond[2]), Xi-1);
          d4 = Math.min((int)((p.data[3] - DimLimits[3].l) / mond[3]), Xi-1);
          CellArray[d1][d2][d3][d4]++;
        }

        /* Create the first bucket */
        NumOfBuckets = 1;

        BucketArray[0].Bounds[0][0] = 0 ;
        BucketArray[0].Bounds[0][1] = Xi-1;
        BucketArray[0].Bounds[1][0] = 0;
        BucketArray[0].Bounds[1][1] = Xi-1;
        BucketArray[0].Bounds[2][0] = 0;
        BucketArray[0].Bounds[2][1] = Xi-1;
        BucketArray[0].Bounds[3][0] = 0;
        BucketArray[0].Bounds[3][1] = Xi-1;

        BucketArray[0].Value = numofpoints;

        b1.Bounds[0][0] = 0 ;
        b1.Bounds[0][1] = Xi-1;
        b1.Bounds[1][0] = 0;
        b1.Bounds[1][1] = Xi-1;
        b1.Bounds[2][0] = 0;
        b1.Bounds[2][1] = Xi-1;
        b1.Bounds[3][0] = 0;
        b1.Bounds[3][1] = Xi-1;

        TotalCells = (double) (Xi*Xi*Xi*Xi);

        GlobalAverage = (double) numofpoints / TotalCells;

        /* Compute the variance */

        d1 = b1.Bounds[0][1] - b1.Bounds[0][0];
        d2 = b1.Bounds[1][1] - b1.Bounds[1][0];
        d3 = b1.Bounds[2][1] - b1.Bounds[2][0];
        d4 = b1.Bounds[3][1] - b1.Bounds[3][0];

        GlobalVariance = 0.0;

        for (i1=0; i1 <= d1 ; i1++)
        {
                for (i2=0; i2 <= d2; i2++)
                {
                        for (i3=0; i3 <= d3; i3++)
                        {
                                for (i4=0; i4 <= d4; i4++)
                                {
                                        t1 = i1 + b1.Bounds[0][0];
                                        t2 = i2 + b1.Bounds[1][0];
                                        t3 = i3 + b1.Bounds[2][0];
                                        t4 = i4 + b1.Bounds[3][0];

                                        GlobalVariance += Math.pow((double) (GlobalAverage - CellArray[t1][t2][t3][t4]), 2.0);
                                }
                        }
                }
        }

        GlobalVariance = GlobalVariance / TotalCells;

        b1.Variance = GlobalVariance;
        BucketArray[0].Variance = GlobalVariance;

        /* now compute the marginal distributions */
        /* first initialize MD */

        for (i=0; i < DIM; i++)
        {
                for (i1=0; i1 < Xi; i1++)
                {
                        BucketArray[0].MD[i][i1] = 0;
                }
        }

        d1 = b1.Bounds[0][1] - b1.Bounds[0][0];
        d2 = b1.Bounds[1][1] - b1.Bounds[1][0];
        d3 = b1.Bounds[2][1] - b1.Bounds[2][0];
        d4 = b1.Bounds[3][1] - b1.Bounds[3][0];

        for (i1=0; i1 <= d1 ; i1++)
        {
                for (i2=0; i2 <= d2; i2++)
                {
                        for (i3=0; i3 <= d3; i3++)
                        {
                                for (i4=0; i4 <= d4; i4++)
                                {
                                        t1 = i1 + b1.Bounds[0][0];
                                        t2 = i2 + b1.Bounds[1][0];
                                        t3 = i3 + b1.Bounds[2][0];
                                        t4 = i4 + b1.Bounds[3][0];

                                        BucketArray[0].MD[0][i1] += CellArray[t1][t2][t3][t4];
                                        BucketArray[0].MD[1][i2] += CellArray[t1][t2][t3][t4];
                                        BucketArray[0].MD[2][i3] += CellArray[t1][t2][t3][t4];
                                        BucketArray[0].MD[3][i4] += CellArray[t1][t2][t3][t4];
                                }
                        }
                }
        }

        /* Now try to split the buckets... */

        NumberOfBuckets = 1;

        while (NumberOfBuckets < NumbOfBins)
        {
                /* first find the bucket to split, the one with the largest gain */
                BuckToSplit = -1;
                SplitDimension = -1;
                BestSplitPoint = -1;
                BestVarDecrease = 0;

                for (i = 0; i < NumberOfBuckets; i++)
                {
                        for (k = 0; k < DIM; k++)   AvgDim[k] = 0.0;

                        d1 = BucketArray[i].Bounds[0][1] - BucketArray[i].Bounds[0][0];
                        d2 = BucketArray[i].Bounds[1][1] - BucketArray[i].Bounds[1][0];
                        d3 = BucketArray[i].Bounds[2][1] - BucketArray[i].Bounds[2][0];
                        d4 = BucketArray[i].Bounds[3][1] - BucketArray[i].Bounds[3][0];

                        /* find the best dimension in Bucket i */
                        AvgDim[0] = BucketArray[i].Value;
                        AvgDim[0] = AvgDim[0]/(d1+1);

                        AvgDim[1] = BucketArray[i].Value;
                        AvgDim[1] = AvgDim[1]/(d2+1);

                        AvgDim[2] = BucketArray[i].Value;
                        AvgDim[2] = AvgDim[2]/(d3+1);

                        AvgDim[3] = BucketArray[i].Value;
                        AvgDim[3] = AvgDim[3]/(d4+1);

                        for (k =0; k < DIM; k++) VarDim[k] = 0.0;

                        for (j = 0; j <= d1; j++)
                        {
                                VarDim[0] += (AvgDim[0] - BucketArray[i].MD[0][j]) * (AvgDim[0] - BucketArray[i].MD[0][j]);
                        }

                        VarDim[0] = VarDim[0]/(d1+1);

                        for (j = 0; j <= d2; j++)
                        {
                                VarDim[1] +=  (AvgDim[1] - BucketArray[i].MD[1][j])*(AvgDim[1] - BucketArray[i].MD[1][j]);
                        }

                        VarDim[1] = VarDim[1]/(d2+1);

                        for (j = 0; j <= d3; j++)
                        {
                                VarDim[2] +=  (AvgDim[2] - BucketArray[i].MD[2][j])*(AvgDim[2] - BucketArray[i].MD[2][j]);
                        }

                        VarDim[2] = VarDim[2]/(d3+1);

                        for (j = 0; j <= d4; j++)
                        {
                                VarDim[3] +=  (AvgDim[3] - BucketArray[i].MD[3][j])*(AvgDim[3] - BucketArray[i].MD[3][j]);
                        }

                        VarDim[3] = VarDim[3]/(d4+1);

                        /* find the dimension with the smallest variance */
                        SmallestVar = -1.0;
                        SmallestDim = -1;

                        for (k=0; k < DIM; k++)
                        {
                                if (VarDim[k] > SmallestVar)
                                {
                                        SmallestVar = VarDim[k];
                                        SmallestDim = k;
                                }
                        }

                        /*OK. Now try to split along the SmallestDim dimension */
                        /* first get the bounds on this dimension */
                        low = BucketArray[i].Bounds[SmallestDim][0];
                        high = BucketArray[i].Bounds[SmallestDim][1];

                        if ((0 == SmallestDim) && ( high - low >= 1))
                        {
                                /* Now compute the best split on that dimension */
                                /* for each split */
                                for (splitp = low; splitp < high; splitp++)
                                {
                                        leftvalue = 0.0;
                                        nleft = 0;
                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= splitp; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        leftvalue += CellArray[i1][i2][i3][i4];
                                                                        nleft++;
                                                                }
                                                        }
                                                }
                                        }

                                        rightvalue = BucketArray[i].Value - leftvalue;
                                        avgleft = leftvalue / nleft;
                                        totaln = 1;

                                        for (k = 0; k < DIM; k++)
                                                totaln *= (BucketArray[i].Bounds[k][1]+1-BucketArray[i].Bounds[k][0]);

                                        nright = totaln - nleft;

                                        avgright = rightvalue / nright;

                                        /* find the new variance */
                                        NewSplitVariance = 0.0;
                                        NewSplitVarianceLeft = 0.0;
                                        NewSplitVarianceRight = 0.0;

                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        if ( i1 <= splitp)
                                                                                NewSplitVarianceLeft += (avgleft - CellArray[i1][i2][i3][i4]) *
                                                                                        (avgleft - CellArray[i1][i2][i3][i4]);
                                                                        else
                                                                                NewSplitVarianceRight += (avgright - CellArray[i1][i2][i3][i4])*
                                                                                        (avgright - CellArray[i1][i2][i3][i4]);
                                                                }
                                                        }
                                                }
                                        }

                                        NewSplitVariance = ((NewSplitVarianceLeft*leftvalue) + (NewSplitVarianceRight*rightvalue)) / totaln;

                                        VarDecrease = BucketArray[i].Value * BucketArray[i].Variance - NewSplitVariance;

                                        if (VarDecrease > BestVarDecrease)
                                        {
                                                SplitDimension = 0;
                                                BestSplitPoint = splitp;
                                                BestVarDecrease = VarDecrease;
                                                BuckToSplit = i;
                                        }
         /* for each split */
                                }
         /* Dimension 0 */
                        }

                        if ((1 == SmallestDim) && ( high - low >= 1))
                        {
                                /* Now compute the best split on that dimension */
                                /* for each split */
                                for (splitp = low; splitp < high; splitp++)
                                {
                                        leftvalue = 0.0;
                                        nleft = 0;
                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= splitp; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        leftvalue += CellArray[i1][i2][i3][i4];
                                                                        nleft++;
                                                                }
                                                        }
                                                }
                                        }

                                        rightvalue = BucketArray[i].Value - leftvalue;
                                        avgleft = leftvalue / nleft;
                                        totaln = 1;

                                        for (k = 0; k < DIM; k++)
                                                totaln *= (BucketArray[i].Bounds[k][1]+1-BucketArray[i].Bounds[k][0]);

                                        nright = totaln - nleft;

                                        avgright = rightvalue / nright;

                                        /* find the new variance */
                                        NewSplitVariance = 0.0;
                                        NewSplitVarianceLeft = 0.0;
                                        NewSplitVarianceRight = 0.0;

                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        if ( i2 <= splitp)
                                                                                NewSplitVarianceLeft += (avgleft - CellArray[i1][i2][i3][i4])*
                                                                                        (avgleft - CellArray[i1][i2][i3][i4]);
                                                                        else
                                                                                NewSplitVarianceRight += (avgright - CellArray[i1][i2][i3][i4])*
                                                                                        (avgright - CellArray[i1][i2][i3][i4]);
                                                                }
                                                        }
                                                }
                                        }

                                        NewSplitVariance = ((NewSplitVarianceLeft*leftvalue) + (NewSplitVarianceRight*rightvalue)) / totaln;

                                        VarDecrease = BucketArray[i].Value * BucketArray[i].Variance - NewSplitVariance;

                                        if (VarDecrease > BestVarDecrease)
                                        {
                                                SplitDimension = 1;
                                                BestSplitPoint = splitp;
                                                BestVarDecrease = VarDecrease;
                                                BuckToSplit = i;
                                        }
         /* for each split */
                                }
         /* Dimension 1 */
                        }

                        if ((2 == SmallestDim) && ( high - low >= 1))
                        {
                                /* Now compute the best split on that dimension */
                                /* for each split */
                                for (splitp = low; splitp < high; splitp++)
                                {
                                        leftvalue = 0.0;
                                        nleft = 0;
                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= splitp; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        leftvalue += CellArray[i1][i2][i3][i4];
                                                                        nleft++;
                                                                }
                                                        }
                                                }
                                        }

                                        rightvalue = BucketArray[i].Value - leftvalue;

                                        avgleft = leftvalue / nleft;

                                        totaln = 1;

                                        for (k = 0; k < DIM; k++)
                                                totaln *= (BucketArray[i].Bounds[k][1]+1-BucketArray[i].Bounds[k][0]);

                                        nright = totaln - nleft;

                                        avgright = rightvalue / nright;

                                        /* find the new variance */
                                        NewSplitVariance = 0.0;
                                        NewSplitVarianceLeft = 0.0;
                                        NewSplitVarianceRight = 0.0;

                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        if ( i3 <= splitp)
                                                                                NewSplitVarianceLeft += (avgleft - CellArray[i1][i2][i3][i4])*
                                                                                        (avgleft - CellArray[i1][i2][i3][i4]);
                                                                        else
                                                                                NewSplitVarianceRight += (avgright - CellArray[i1][i2][i3][i4])*
                                                                                        (avgright - CellArray[i1][i2][i3][i4]);
                                                                }
                                                        }
                                                }
                                        }

                                        NewSplitVariance = ((NewSplitVarianceLeft*leftvalue) + (NewSplitVarianceRight*rightvalue)) / totaln;

                                        VarDecrease = BucketArray[i].Value * BucketArray[i].Variance - NewSplitVariance;

                                        if (VarDecrease > BestVarDecrease)
                                        {
                                                SplitDimension = 2;
                                                BestSplitPoint = splitp;
                                                BestVarDecrease = VarDecrease;
                                                BuckToSplit = i;
                                        }
         /* for each split */
                                }
         /* Dimension 2 */
                        }

                        if ((3 == SmallestDim) && ( high - low >= 1))
                        {
                                /* Now compute the best split on that dimension */
                                /* for each split */
                                for (splitp = low; splitp < high; splitp++)
                                {
                                        leftvalue = 0.0;
                                        nleft = 0;
                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= splitp; i4++)
                                                                {
                                                                        leftvalue += CellArray[i1][i2][i3][i4];
                                                                        nleft++;
                                                                }
                                                        }
                                                }
                                        }

                                        rightvalue = BucketArray[i].Value - leftvalue;
                                        avgleft = leftvalue / nleft;
                                        totaln = 1;

                                        for (k = 0; k < DIM; k++)
                                                totaln *= (BucketArray[i].Bounds[k][1]+1-BucketArray[i].Bounds[k][0]);

                                        nright = totaln - nleft;

                                        avgright = rightvalue / nright;

                                        /* find the new variance */
                                        NewSplitVariance = 0.0;
                                        NewSplitVarianceLeft = 0.0;
                                        NewSplitVarianceRight = 0.0;

                                        for (i1=BucketArray[i].Bounds[0][0]; i1 <= BucketArray[i].Bounds[0][1]; i1++)
                                        {
                                                for (i2=BucketArray[i].Bounds[1][0]; i2 <= BucketArray[i].Bounds[1][1]; i2++)
                                                {
                                                        for (i3=BucketArray[i].Bounds[2][0]; i3 <= BucketArray[i].Bounds[2][1]; i3++)
                                                        {
                                                                for (i4=BucketArray[i].Bounds[3][0]; i4 <= BucketArray[i].Bounds[3][1]; i4++)
                                                                {
                                                                        if ( i4 <= splitp)
                                                                                NewSplitVarianceLeft += (avgleft - CellArray[i1][i2][i3][i4])*
                                                                                        (avgleft - CellArray[i1][i2][i3][i4]);
                                                                        else
                                                                                NewSplitVarianceRight += (avgright - CellArray[i1][i2][i3][i4])*
                                                                                        (avgright - CellArray[i1][i2][i3][i4]);
                                                                }
                                                        }
                                                }
                                        }

                                        NewSplitVariance = ((NewSplitVarianceLeft*leftvalue) + (NewSplitVarianceRight*rightvalue)) / totaln;
                                        VarDecrease = BucketArray[i].Value * BucketArray[i].Variance - NewSplitVariance;

                                        if (VarDecrease > BestVarDecrease)
                                        {
                                                SplitDimension = 3;
                                                BestSplitPoint = splitp;
                                                BestVarDecrease = VarDecrease;
                                                BuckToSplit = i;
                                        }
         /* for each split */
                                }
         /* Dimension 3 */
                        }
         /* for each bucket */
                }

                /* Now we now the bucket, dimension, splitpoint and variance decrease */
                /* We are ready to split the bucket into two by adding a new one and updating the values of the old one */
                if (BuckToSplit == -1 ) break;

                NumberOfBuckets++;

                /* copy the bounds of the old to the new  bucket */
                BucketArray[NumberOfBuckets-1].Bounds[0][0] = BucketArray[BuckToSplit].Bounds[0][0];
                BucketArray[NumberOfBuckets-1].Bounds[0][1] = BucketArray[BuckToSplit].Bounds[0][1];
                BucketArray[NumberOfBuckets-1].Bounds[1][0] = BucketArray[BuckToSplit].Bounds[1][0];
                BucketArray[NumberOfBuckets-1].Bounds[1][1] = BucketArray[BuckToSplit].Bounds[1][1];
                BucketArray[NumberOfBuckets-1].Bounds[2][0] = BucketArray[BuckToSplit].Bounds[2][0];
                BucketArray[NumberOfBuckets-1].Bounds[2][1] = BucketArray[BuckToSplit].Bounds[2][1];
                BucketArray[NumberOfBuckets-1].Bounds[3][0] = BucketArray[BuckToSplit].Bounds[3][0];
                BucketArray[NumberOfBuckets-1].Bounds[3][1] = BucketArray[BuckToSplit].Bounds[3][1];

                /* update the bound for the split dimension */
                BucketArray[NumberOfBuckets-1].Bounds[SplitDimension][0] = BestSplitPoint+1;

                /* Update Bounds of the old bucket */
                BucketArray[BuckToSplit].Bounds[SplitDimension][1] = BestSplitPoint;

                /* Now we need to update the MD, Value and Variance */
                /* first initialize MD and the update */
                for (i=0; i < DIM; i++)
                {
                        for (i1=0; i1 < Xi; i1++)
                        {
                                BucketArray[BuckToSplit].MD[i][i1] = 0;
                        }
                }

                d1 = BucketArray[BuckToSplit].Bounds[0][1] - BucketArray[BuckToSplit].Bounds[0][0];
                d2 = BucketArray[BuckToSplit].Bounds[1][1] - BucketArray[BuckToSplit].Bounds[1][0];
                d3 = BucketArray[BuckToSplit].Bounds[2][1] - BucketArray[BuckToSplit].Bounds[2][0];
                d4 = BucketArray[BuckToSplit].Bounds[3][1] - BucketArray[BuckToSplit].Bounds[3][0];

                for (i1=0; i1 <= d1 ; i1++)
                {
                        for (i2=0; i2 <= d2; i2++)
                        {
                                for (i3=0; i3 <= d3; i3++)
                                {
                                        for (i4=0; i4 <= d4; i4++)
                                        {
                                                t1 = i1 + BucketArray[BuckToSplit].Bounds[0][0];
                                                t2 = i2 + BucketArray[BuckToSplit].Bounds[1][0];
                                                t3 = i3 + BucketArray[BuckToSplit].Bounds[2][0];
                                                t4 = i4 + BucketArray[BuckToSplit].Bounds[3][0];

                                                BucketArray[BuckToSplit].MD[0][i1] += CellArray[t1][t2][t3][t4];
                                                BucketArray[BuckToSplit].MD[1][i2] += CellArray[t1][t2][t3][t4];
                                                BucketArray[BuckToSplit].MD[2][i3] += CellArray[t1][t2][t3][t4];
                                                BucketArray[BuckToSplit].MD[3][i4] += CellArray[t1][t2][t3][t4];
                                        }
                                }
                        }
                }

                /* Do the same for the new bucket.. */
                for (i=0; i < DIM; i++)
                {
                        for (i1=0; i1 < Xi; i1++)
                        {
                                BucketArray[NumberOfBuckets-1].MD[i][i1] = 0;
                        }
                }

                d1 = BucketArray[NumberOfBuckets-1].Bounds[0][1] - BucketArray[NumberOfBuckets-1].Bounds[0][0];
                d2 = BucketArray[NumberOfBuckets-1].Bounds[1][1] - BucketArray[NumberOfBuckets-1].Bounds[1][0];
                d3 = BucketArray[NumberOfBuckets-1].Bounds[2][1] - BucketArray[NumberOfBuckets-1].Bounds[2][0];
                d4 = BucketArray[NumberOfBuckets-1].Bounds[3][1] - BucketArray[NumberOfBuckets-1].Bounds[3][0];

                for (i1=0; i1 <= d1 ; i1++)
                {
                        for (i2=0; i2 <= d2; i2++)
                        {
                                for (i3=0; i3 <= d3; i3++)
                                {
                                        for (i4=0; i4 <= d4; i4++)
                                        {
                                                t1 = i1 + BucketArray[NumberOfBuckets-1].Bounds[0][0];
                                                t2 = i2 + BucketArray[NumberOfBuckets-1].Bounds[1][0];
                                                t3 = i3 + BucketArray[NumberOfBuckets-1].Bounds[2][0];
                                                t4 = i4 + BucketArray[NumberOfBuckets-1].Bounds[3][0];

                                                BucketArray[NumberOfBuckets-1].MD[0][i1] += CellArray[t1][t2][t3][t4];
                                                BucketArray[NumberOfBuckets-1].MD[1][i2] += CellArray[t1][t2][t3][t4];
                                                BucketArray[NumberOfBuckets-1].MD[2][i3] += CellArray[t1][t2][t3][t4];
                                                BucketArray[NumberOfBuckets-1].MD[3][i4] += CellArray[t1][t2][t3][t4];
                                        }
                                }
                        }
                }

                /* Now compute the new Values */
                NewBValue = 0;
                NewCellNumb = 0;

                for (i1=BucketArray[NumberOfBuckets-1].Bounds[0][0]; i1 <= BucketArray[NumberOfBuckets-1].Bounds[0][1]; i1++)
                {
                        for (i2=BucketArray[NumberOfBuckets-1].Bounds[1][0]; i2 <= BucketArray[NumberOfBuckets-1].Bounds[1][1]; i2++)
                        {
                                for (i3=BucketArray[NumberOfBuckets-1].Bounds[2][0]; i3 <= BucketArray[NumberOfBuckets-1].Bounds[2][1]; i3++)
                                {
                                        for (i4=BucketArray[NumberOfBuckets-1].Bounds[3][0]; i4 <= BucketArray[NumberOfBuckets-1].Bounds[3][1]; i4++)
                                        {
                                                NewBValue += CellArray[i1][i2][i3][i4];
                                                NewCellNumb++;
                                        }
                                }
                        }
                }

                BucketArray[NumberOfBuckets-1].Value  = NewBValue;
                BucketArray[BuckToSplit].Value = BucketArray[BuckToSplit].Value - NewBValue;

                /* Now compute the new total Variance for each bucket */
                AvgNewB = (double) NewBValue / (double) NewCellNumb;

                /* find the new number of Cells in the old bucket */
                OldCellNum = 1;
                for (k = 0; k < DIM; k++)
                        OldCellNum  *= (BucketArray[BuckToSplit].Bounds[k][1]+1-BucketArray[BuckToSplit].Bounds[k][0]);

                AvgOldB = (double) BucketArray[BuckToSplit].Value / (double) OldCellNum;

                /* find the new variance */
                NewBuckVariance = 0.0;

                d1 = BucketArray[NumberOfBuckets-1].Bounds[0][1] - BucketArray[NumberOfBuckets-1].Bounds[0][0];
                d2 = BucketArray[NumberOfBuckets-1].Bounds[1][1] - BucketArray[NumberOfBuckets-1].Bounds[1][0];
                d3 = BucketArray[NumberOfBuckets-1].Bounds[2][1] - BucketArray[NumberOfBuckets-1].Bounds[2][0];
                d4 = BucketArray[NumberOfBuckets-1].Bounds[3][1] - BucketArray[NumberOfBuckets-1].Bounds[3][0];

                for (i1=0; i1 <= d1 ; i1++)
                {
                        for (i2=0; i2 <= d2; i2++)
                        {
                                for (i3=0; i3 <= d3; i3++)
                                {
                                        for (i4=0; i4 <= d4; i4++)
                                        {
                                                t1 = i1 + BucketArray[NumberOfBuckets-1].Bounds[0][0];
                                                t2 = i2 + BucketArray[NumberOfBuckets-1].Bounds[1][0];
                                                t3 = i3 + BucketArray[NumberOfBuckets-1].Bounds[2][0];
                                                t4 = i4 + BucketArray[NumberOfBuckets-1].Bounds[3][0];

                                                NewBuckVariance += Math.pow((double) (AvgNewB - CellArray[t1][t2][t3][t4]), 2.0);
                                        }
                                }
                        }
                }

                NewBuckVariance = NewBuckVariance / (double) NewCellNumb;
                BucketArray[NumberOfBuckets-1].Variance = NewBuckVariance;

                /* DO the same for the old one */
                OldBuckVariance = 0.0;

                d1 = BucketArray[BuckToSplit].Bounds[0][1] - BucketArray[BuckToSplit].Bounds[0][0];
                d2 = BucketArray[BuckToSplit].Bounds[1][1] - BucketArray[BuckToSplit].Bounds[1][0];
                d3 = BucketArray[BuckToSplit].Bounds[2][1] - BucketArray[BuckToSplit].Bounds[2][0];
                d4 = BucketArray[BuckToSplit].Bounds[3][1] - BucketArray[BuckToSplit].Bounds[3][0];

                for (i1=0; i1 <= d1 ; i1++)
                {
                        for (i2=0; i2 <= d2; i2++)
                        {
                                for (i3=0; i3 <= d3; i3++)
                                {
                                        for (i4=0; i4 <= d4; i4++)
                                        {
                                                t1 = i1 + BucketArray[BuckToSplit].Bounds[0][0];
                                                t2 = i2 + BucketArray[BuckToSplit].Bounds[1][0];
                                                t3 = i3 + BucketArray[BuckToSplit].Bounds[2][0];
                                                t4 = i4 + BucketArray[BuckToSplit].Bounds[3][0];

                                                OldBuckVariance += Math.pow((double) (AvgOldB - CellArray[t1][t2][t3][t4]), 2.0);
                                        }
                                }
                        }
                }

                OldBuckVariance = OldBuckVariance / (double) OldCellNum;
                BucketArray[BuckToSplit].Variance = OldBuckVariance;

                /* Split done! */
         /* while not finished */
        }

        Vector ret = new Vector();
        GlobalVariance = 0.0;

        for (i = 0; i < NumberOfBuckets; i++)
        {
                Bucket4D b = new Bucket4D();

                b.low[0] = BucketArray[i].Bounds[0][0] * mond[0] + DimLimits[0].l;
                b.high[0] = (BucketArray[i].Bounds[0][1] + 1) * mond[0] + DimLimits[0].l;
                b.low[1] = BucketArray[i].Bounds[1][0] * mond[1] + DimLimits[1].l;
                b.high[1] = (BucketArray[i].Bounds[1][1] + 1) * mond[1] + DimLimits[1].l;
                b.low[2] = BucketArray[i].Bounds[2][0] * mond[2] + DimLimits[2].l;
                b.high[2] = (BucketArray[i].Bounds[2][1] + 1) * mond[2] + DimLimits[2].l;
                b.low[3] = BucketArray[i].Bounds[3][0] * mond[3] + DimLimits[3].l;
                b.high[3] = (BucketArray[i].Bounds[3][1] + 1) * mond[3] + DimLimits[3].l;
                b.count = BucketArray[i].Value;
                GlobalVariance += BucketArray[i].Value * BucketArray[i].Variance;

                ret.add(b);

                //cerr << b.low[0] << " " << b.high[0] << " " << b.low[1] << " " << b.high[1] << " " << b.count << endl;
        }

        System.out.println("Average variance: " + GlobalVariance + "(variance) / "
                   + NumberOfBuckets  + " (buckets) = " + GlobalVariance / (double) NumberOfBuckets + " per bucket.");

        return ret;
  }

  public static long checkHistogram(BufferTuple tuple, Vector histogram)
  {
    long count = -1;
    if (StreamJoin.NUM_ATTRIBUTE == 4) {
      int size = histogram.size();
      for (int i = 0; i < size; i++) {
        Bucket4D bucket = (Bucket4D) histogram.elementAt(i);
        boolean inside = true;
        for (int j = 0; j < 4; j++) {
          if (bucket.low[j] > tuple.value[j] || bucket.high[j] < tuple.value[j]) {
            inside = false;
            break;
          }
        }
        if (inside) {
          count = bucket.count;
          break;
        }
      }
    }
    return count;
  }

  public static boolean updateHistogram(BufferTuple tuple, Vector histogram)
  {
    boolean found = false;
    if (StreamJoin.NUM_ATTRIBUTE == 4)
    {
      int size = histogram.size();
      for (int i = 0; i < size; i++)
      {
        Bucket4D bucket = (Bucket4D) histogram.elementAt(i);
        boolean inside = true;
        for (int j = 0; j < 4; j++)
        {
          if (bucket.low[j] > tuple.value[j] || bucket.high[j] < tuple.value[j])
          {
            inside = false;
            break;
          }
        }
        if (inside)
        {
          bucket.count++;
          found = true;
          break;
        }
      }
    }
    return found;
  }
}
