package StreamSimulator.Stream;

/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */
import StreamSimulator.*;
import java.util.*;
import java.util.Vector;
import StreamSimulator.Buffer.*;
import StreamSimulator.Clock.*;
import StreamSimulator.MinSkewPacket.*;
import StreamSimulator.MinSkewPacket.MinSkew2D.*;
import StreamSimulator.MinSkewPacket.MinSkew4D.*;
import StreamSimulator.StreamJoin.*;

public class Topology implements ClockListener {

  public long[][] joinResults;
  public Vector streams;
  public JoinBuffer[] buffer;
  public int bufferSize;
  public int currentBufferSize;
  public int bufferAlgo;
  public int windowSize;
  public boolean initialDone;
  public HashMap statistics;
  public int done;
  public int numStream;
  public Vector[] streamData;
  public Vector[] histograms;
  public int[][] arrivalRate;
  public double[] avgArrivalRate;
  public boolean useHistogram;
  public long rebuild_interval;
  public int gridSize;
  public int numOfBucket;
  public long warm_up;
  public boolean seperate_buffer;
  public int numBuffer;
  public long timestamp;

  public static int NORMAL_SHEDDING = 0;
  public static int RANDOM_SHEDDING = 1;
  public static int PROBABILITY_SHEDDING = 2;
  public static int AGING_SHEDDING = 3;
  public static int STANFORD_SHEDDING = 4;  
  public static int OPTIMAL_SHEDDING = 5;


  public Topology(int bufferAlgo, int bufferSize, int windowSize, int num_streams,long warmUp,
                  boolean useHistogram, boolean seperate_buffer, int[] buffers, int genType, int totalTime) {
    this.bufferSize = bufferSize;
    this.windowSize = windowSize;
    this.bufferAlgo = bufferAlgo;
    this.numStream = num_streams;
    this.useHistogram = useHistogram;
    this.warm_up = warmUp;
    this.initialDone = false;
    this.rebuild_interval = 0;
    this.seperate_buffer = seperate_buffer;
    if (seperate_buffer == false)
    {
      buffer = new JoinBuffer[1];
      this.numBuffer = 1;
      if (bufferAlgo == RANDOM_SHEDDING)
        buffer[0] = new RandomSheddingBuffer(buffers[0], this.numStream);
      else if (bufferAlgo == PROBABILITY_SHEDDING)
        buffer[0] = new ProbabilitySheddingBuffer(buffers[0], this.numStream);
      else if (bufferAlgo == AGING_SHEDDING)
        buffer[0] = new AgingBuffer(buffers[0], 1.0,this.seperate_buffer, this.numStream);
      else if (bufferAlgo == NORMAL_SHEDDING)
        buffer[0] = new JoinBuffer(buffers[0], this.numStream);
      else if (bufferAlgo == STANFORD_SHEDDING)
      	buffer[0] = new StanfordBuffer(buffers[0], this.numStream, 1);
    }
    else
    {
       buffer = new JoinBuffer[this.numStream];
       this.numBuffer = this.numStream;
       for (int i=0;i<this.numStream;i++)
       {
         String file = "bakdata"+i;
         if (bufferAlgo == RANDOM_SHEDDING)
           buffer[i] = new RandomSheddingBuffer(buffers[i],this.numStream);
         else if (bufferAlgo == PROBABILITY_SHEDDING)
           buffer[i] = new ProbabilitySheddingBuffer(buffers[i], this.numStream);
         else if (bufferAlgo == AGING_SHEDDING)
           buffer[i] = new AgingBuffer(buffers[i], 1.0,this.seperate_buffer, this.numStream);
         else if (bufferAlgo == NORMAL_SHEDDING)
           buffer[i] = new JoinBuffer(buffers[i], this.numStream);
         else if (bufferAlgo == OPTIMAL_SHEDDING)
           buffer[i] = new OptimalBuffer(buffers[i], this.numStream, file,genType);
         else if (bufferAlgo == STANFORD_SHEDDING)
           buffer[i] = new StanfordBuffer(buffers[i], this.numStream, i+1);
       }
    }
    streams = new Vector();
    statistics = new HashMap();
    done = 0;
    joinResults = new long[numStream][StreamJoin.NUM_ATTRIBUTE];
    if (this.useHistogram == true)
    {
      streamData = new Vector[numStream];
      histograms = new Vector[numStream];
      for (int i = 0; i < this.numStream; i++) {
        this.streamData[i] = new Vector();
        this.histograms[i] = new Vector();
      }
    }
    for (int i=0;i<this.numStream;i++)
      for (int j=0;j<StreamJoin.NUM_ATTRIBUTE;j++)
        joinResults[i][j] = 0;
    arrivalRate = new int[totalTime][numStream];
    for (int i=0;i<totalTime;i++)
      for (int j=0;j<numStream;j++)
        arrivalRate[i][j] = 0;
    avgArrivalRate = new double[numStream];
    for (int j=0;j<numStream;j++)
        avgArrivalRate[j] = 0.0;
  }

  public void setTime(long t)
  {
  	this.timestamp = t;
     for (int i=0;i<this.numBuffer;i++)
     {
       buffer[i].validateBuffer(t, windowSize);
       if (bufferAlgo == AGING_SHEDDING)
         buffer[i].aging(t);
       if (bufferAlgo == STANFORD_SHEDDING)
         buffer[i].aging(t);
       if (bufferAlgo == OPTIMAL_SHEDDING);
         buffer[i].moveToNextTuple(t-StreamJoin.UNIT,windowSize,StreamJoin.UNIT);
     }

    if (this.useHistogram == true)
    {
      if ((t-this.warm_up)%this.rebuild_interval==0 && t>=this.warm_up)
        this.buildHistogram();
    }
    int tmp = (int)((t-Clock.stime)/StreamJoin.UNIT)%this.windowSize;
    
    if (tmp==0)
    {
       System.out.println("done with time "+t);
       for (int i=0;i<this.numBuffer;i++)
         System.out.println("buffer size for each buffer\n" + 
        this.buffer[i].currentSize+"/"+this.buffer[i].bufferSize+" "+this.buffer[i].buffer.size());
       long total = 0;

       for (int i=0;i<this.numBuffer;i++)
       {
         int[] streamCount = this.buffer[i].countStreamSize();
         for (int j=0;j<this.numStream;j++)
           System.out.print("common buffer " + j + " " + streamCount[j]+" ");
         System.out.print("\n");
       }
       //int unit = (int)((t-Clock.stime)/StreamJoin.UNIT-1);
       //this.getAvgArrivalRate(this.windowSize);
       //for (int i=0;i<this.numStream;i++)
       	//System.out.println("arrival rate for stream "+i+" "+this.avgArrivalRate[i]);
       for (int i=0;i<this.numStream;i++)
       {  for (int j=0;j<StreamJoin.NUM_ATTRIBUTE;j++)
          {
            total += this.joinResults[i][j];
            System.out.print("join results for stream " + i + " " + joinResults[i][j]+" ");
          }
          System.out.print("\n");
       }
       System.out.println(total);
       /*this.buffer[0].printBuffer();
       System.out.println("fffffffff");
       this.buffer[1].printBuffer();*/
    }
  }

  public void SetHistogram(long rebuild_interval, int grid_size, int buckets)
  {
    this.useHistogram = true;
    this.rebuild_interval = rebuild_interval;
    this.gridSize = grid_size;
    this.numOfBucket = buckets;
    for (int i=0;i<this.numBuffer;i++)
      this.buffer[i].use_histogram = true;
  }

  public void buildHistogram()
  {
    if (StreamJoin.NUM_ATTRIBUTE ==2)
    {
      MinSkew2D MinSkew2D = new MinSkew2D();
      for (int i=0;i<this.numStream;i++)
      {
        //System.out.println(this.streamData[i].size());
        this.histograms[i] = MinSkew2D.createHistogram(this.streamData[i],this.gridSize,this.numOfBucket);
      }
    }
    else if (StreamJoin.NUM_ATTRIBUTE ==4)
    {
      MinSkew4D MinSkew4D = new MinSkew4D();
      for (int i=0;i<this.numStream;i++)
      {
        this.histograms[i] = MinSkew4D.createHistogram(this.streamData[i],this.gridSize,this.numOfBucket);
      }
    }
  }
  
  public void getAvgArrivalRate(int historyLength)
  {  
  		double[] total=new double[numStream];
  		for (int i=0;i<numStream;i++)
  			total[i] = 0.0;
  		int unit = (int)((this.timestamp-Clock.stime)/StreamJoin.UNIT-1);
  		//System.out.println(unit+" "+historyLength);
  		for (int j=unit; j>unit-historyLength+1;j--)
  		{		
  			for (int i=0;i<numStream;i++)
	  			total[i]+=this.arrivalRate[j][i];
  		}
  		for (int i=0;i<numStream;i++)
  			avgArrivalRate[i]=total[i]/historyLength;	
  
  }
}
