package StreamSimulator.StreamGenerator;
import java.util.Vector;
import java.util.Random;
import java.lang.Math;
import StreamSimulator.Buffer.*;

/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */

public class GaussianGenerator extends StreamGenerator {

  //private int min, max;
  private float mean;
  private double sigma;
  private Random rand;
  private long seed;

  public GaussianGenerator(float mean, double sigma, long seed) {
    super(seed);
    this.seed = seed;
    this.mean = mean;
    this.sigma = sigma;
    this.rand = new Random(seed);
  }

  public Vector nextData()
  {
    double dataTemp = nextGaussian(this.mean,this.sigma);
    double floatpart = dataTemp-Math.floor(dataTemp);
    int data=0;
    if (floatpart>=0.5)
      data = (int)Math.ceil(dataTemp);
    else
      data = (int)Math.floor(dataTemp);
    Vector v = new Vector();
    v.add(new Integer(data));
    return v;
  }

  // See Knuth, ACP, Section 3.4.1 Algorithm C.
  synchronized public double nextGaussian(float mean, double sigma)
  {
    boolean haveNextNextGaussian = false;
    double nextNextGaussian=0.0;
    if (haveNextNextGaussian)
    {
      haveNextNextGaussian = false;
      return nextNextGaussian;
    }
    else
    {
      double v1, v2, s;
      do {
         v1 = 2 * rand.nextDouble() - 1; // between -1 and 1
         v2 = 2 * rand.nextDouble() - 1; // between -1 and 1
         s = v1 * v1 + v2 * v2;
      } while (s >= 1 || s == 0);
      double multiplier = Math.sqrt(-2 * Math.log(s)/s);
      nextNextGaussian = v2 * multiplier*sigma+mean;
      haveNextNextGaussian = true;
      return (v1 * multiplier*sigma+mean);
     }
   }
}
