package StreamSimulator.StreamGenerator;

/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */
import StreamSimulator.Buffer.*;
import java.util.Vector;
import java.util.HashMap;
import java.lang.String;

import StreamSimulator.StreamGenerator.TimeCorrelation.*;
import StreamSimulator.StreamJoin.StreamJoin;

import java.io.*;

public class Generator {
  public long timeStamp;
  public Vector streamData;
  public HashMap statistics;
  protected long seed;
  public int streamID;
  public int sameAttrGen;
  public Vector streamGen;
  public static long ATTR_SEED_STEP = 10000L;
  public int domainSize;
  public int domainMinVal;
  public double skewness;
  public String file;
  public int stream_id;
  public int genType;
  public boolean keep_statistics;
  public boolean write_toFile;
  public FileWriter outputFile;


  public static int UNIFORM = 0;
  public static int GAUSSIAN = 1;
  public static int PARETO = 2;
  public static int TGGAUSSIAN = 3;
  public static int TGRANDGAUSSIAN = 4;
  public static int TGRANDUNIFORM = 5;
  public static int WC = 6;
  public static int DAP = 7;
  public static int FILEGEN = 8;
  public static int STOCK = 9;
  public static int GISMO = 10;
  public static int WEATHER = 11;
  public static int REVERSEPARETO = 12;

  public Generator(long _seed, int _sameAttrGen, int _genType, int _domainMinVal,
         int _domainSize, double _skewness,double _sigma1, double _sigma2, int _stream_id,
         int _timeGenType, int _totalData, int _meanStep, int _shift, String _file,
         boolean write_toFile, int range){
    timeStamp = 0;
    this.seed = _seed;
    this.sameAttrGen  = _sameAttrGen;
    this.domainMinVal = _domainMinVal;
    this.domainSize = _domainSize;
    this.skewness = _skewness;
    this.streamID = _stream_id;
    this.file = _file;
    this.write_toFile = write_toFile;
    this.genType = _genType;
    this.keep_statistics = false;
    streamGen = new Vector();
    int count_sameAttr = 0;
    if (this.write_toFile == true)
    {
      try
      {
        this.outputFile = new FileWriter(this.file);
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
    if (genType!= DAP && genType!= WC && genType!=FILEGEN && 
    	genType != GISMO && genType != STOCK && genType != WEATHER)
    {
      for (int i = 0; i < StreamJoin.NUM_ATTRIBUTE; i++)
      {

        if (genType == UNIFORM)
        {
          streamGen.add(new UniformGenerator(this.domainMinVal,
                                             this.domainMinVal +
                                             this.domainSize, seed));
        }
        else if (genType == GAUSSIAN)
        {
            int max = this.domainMinVal + this.domainSize;
            int mean = (int) Math.ceil( (this.domainMinVal + max) / 2);
            streamGen.add(new GaussianGenerator(mean, _sigma1, seed));
        }
        else if (genType == PARETO)
            streamGen.add(new ParetoGenerator(this.domainMinVal,
                                               this.domainMinVal +
                                               this.domainSize,
                                               skewness, seed));
        else if (genType == REVERSEPARETO)
            streamGen.add(new ReverseParetoGenerator(this.domainMinVal,
                                               this.domainMinVal +
                                               this.domainSize,
                                               skewness, seed));

        else if (genType == TGGAUSSIAN)
        {
          GaussianTG tg = new GaussianTG(_timeGenType, _totalData, domainMinVal,
                                       domainSize, seed, _sigma2, skewness,
                                       _meanStep, _shift, streamID);
          tg.generateData();
          tg.shiftData();
          streamGen.add(tg);
        }
        else if (genType == TGRANDGAUSSIAN)
        {
          RandGaussianTG tg = new RandGaussianTG(_timeGenType, _totalData, domainMinVal,
                                         domainSize, seed, _sigma2, skewness,
                                         _meanStep, _shift,streamID);
          tg.generateData();
          tg.shiftData();
          streamGen.add(tg);
        }
        else if (genType == TGRANDUNIFORM)
        {
          RandUniformTG tg = new RandUniformTG(_timeGenType, _totalData,
                                                 domainMinVal,
                                                 domainSize, seed, _sigma2,
                                                 skewness,
                                                 _meanStep, _shift, range,streamID);
          //tg.generateData();
          //tg.generateMean();
          tg.getData();
          tg.getMean();
          tg.shiftDataForWindows();
          streamGen.add(tg);
        }
        seed+=ATTR_SEED_STEP;
        count_sameAttr++;
        if (count_sameAttr>=this.sameAttrGen)
        {
          if (genType !=TGGAUSSIAN)
            genType = (genType+1)%4;
          else
            _timeGenType = (_timeGenType+1)%3;
        }
      }
    }
    else
    {
      if (genType == WC)
        streamGen.add(new WCUPGenerator(file,seed));
      else if(genType == GISMO)
      	streamGen.add(new GismoGenerator(file,seed));
      else if(genType == STOCK)
      	streamGen.add(new StockDataGenerator(file,seed));
      else if(genType == WEATHER)
      	streamGen.add(new WeatherDataGenerator(file,seed));
      else if (genType == DAP)
        streamGen.add(new DAPGenerator(file,seed));
      else if (genType == FILEGEN)
        streamGen.add(new DataFileGenerator(file,seed));
    }
  }

  public void setTime (long t)
  {
    this.timeStamp = t;
    int size = streamGen.size();
    for (int i=0;i<size;i++)
    {
      StreamGenerator gen  = (StreamGenerator)streamGen.elementAt(i);
      gen.timeStamp = t;
    }
  }


  public BufferTuple nextTuple()
  {
    int[] attr_data = new int[StreamJoin.NUM_ATTRIBUTE];
    Vector data = new Vector();
    for (int i=0;i<StreamJoin.NUM_ATTRIBUTE;i++)
      attr_data[i]=0;
    if (genType == WC || genType == DAP || genType == FILEGEN || 
    	genType == GISMO || genType == STOCK || genType == WEATHER)
    {
      StreamGenerator gen = (StreamGenerator) streamGen.elementAt(0);
      data = gen.nextData();
      if (data == null)
        return null;
      int dataSize = data.size();
      for (int j=0;j<dataSize;j++)
         attr_data[j] =  ((Integer)data.elementAt(j)).intValue();
    }
    else
    {
      int size = streamGen.size();
      for (int i= 0;i<size;i++)
      {
        StreamGenerator gen = (StreamGenerator) streamGen.elementAt(i);
        data = gen.nextData();
        attr_data[i] = ((Integer)data.elementAt(0)).intValue();
      }
    }
    BufferTuple tuple = this.updateStatistics(attr_data);
    if (this.write_toFile == true)
    {
      String output = tuple.timestamp+",";
      for (int i = 0;i<StreamJoin.NUM_ATTRIBUTE;i++)
        output=output+tuple.value[i]+",";
      output+="\n";
      try
      {
        this.outputFile.write(output);
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
    return tuple;
  }


  protected synchronized BufferTuple updateStatistics(int[] data)
  {
    BufferTuple tuple = new BufferTuple(this.timeStamp,0.0,this.streamID,data);
    if (this.keep_statistics)
    {
      for (int i = 0; i < StreamJoin.NUM_ATTRIBUTE; i++)
      {
        String key = this.streamID + ":" +i+":"+data[i];
        if (statistics.isEmpty())
          statistics.put(key, new Long(1));
        else
        {
          if (statistics.containsKey(key))
          {
            Long val = (Long) statistics.get(key);
            statistics.put(key, new Long(val.intValue() + 1));
          }
          else
            statistics.put(key, new Long(1));
        }
      }
      String keyTotal = this.streamID+":total";
      if (statistics.isEmpty())
        statistics.put(keyTotal, new Long(1));
      else
      {
        if (statistics.containsKey(keyTotal)) {
          Long val = (Long) statistics.get(keyTotal);
          statistics.put(keyTotal, new Long(val.intValue() + 1));
        }
        else
          statistics.put(keyTotal, new Long(1));
      }
    }
    return tuple;
  }

}
