package StreamSimulator.StreamGenerator.TimeCorrelation;

/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */

import java.util.*;
import StreamSimulator.Buffer.*;
import java.lang.Math;

public class GaussianTG extends TimeDimensionGenerator {

  private Vector usedPos;
  private double sigma;
  private Random posRand;
  public GaussianTG( int genType, int totalData, int domainMinVal,
  int domainSize, long seed, double sigma, double skewness,int meanStep,int shift, int streamID) {
    super(genType,totalData, domainMinVal,domainSize, seed, sigma, skewness,
          meanStep, shift, streamID);
    this.meanStep = meanStep;
    this.sigma = sigma;
    this.usedPos = new Vector();
    posRand = new Random(seed+43210L);
  }

  public void shiftData ()
  {
     Comparator1 comparator = new Comparator1();
     Vector entrySet = new Vector(statistics.entrySet());
     Collections.sort(entrySet,comparator);
     Iterator entrySetIte = entrySet.iterator();
     while(entrySetIte.hasNext())
     {
        Map.Entry e = (Map.Entry)entrySetIte.next();
        Integer key = (Integer)e.getKey();
        Integer value = (Integer)e.getValue();
        //System.out.println("Going to generate :"+key.intValue()+" "+value.intValue());
        genOneGroup(key,value,shift);
     }
  }


  private int genOneGroup(Integer key, Integer value, int shift)
  {
    int count = 0;
    int usedPosSize = usedPos.size();
    int freq = value.intValue();
    int val = key.intValue();
    double mean = freq / 2.0 + usedPosSize;
    while (count < freq)
    {
      int pos;
      do
      {   pos = ((int)this.nextGaussian(mean+shift, sigma))%this.totalData;
      }while (pos<0);
      boolean direction = true;
      if (pos>=mean)
        direction = true;
      else
        direction = false;
      while (usedPos.contains(new Integer(pos)))
      {
        int posStep = posRand.nextInt(this.meanStep)+1;
        if (direction)
          pos = (pos+posStep)%this.totalData;
        else
        {  pos = pos-posStep;
           if (pos<0)
             pos= this.totalData-1;
           pos = pos%this.totalData;
        }
      }
      usedPos.add(new Integer(pos));
      this.streamData[pos] = val;
      count++;
    }
    return count;
  }

  synchronized private double nextGaussian(double mean, double sigma)
  {
    boolean haveNextNextGaussian = false;
    double nextNextGaussian = 0.0;
    if (haveNextNextGaussian) {
      haveNextNextGaussian = false;
      return nextNextGaussian;
    }
    else
    {
      double v1, v2, s;
      do {
        v1 = 2 * rand.nextDouble() - 1; // between -1 and 1
        v2 = 2 * rand.nextDouble() - 1; // between -1 and 1
        s = v1 * v1 + v2 * v2;
      }
      while (s >= 1 || s == 0);
      double multiplier = Math.sqrt( -2 * Math.log(s) / s);
      nextNextGaussian = v2 * multiplier * sigma + mean;
      haveNextNextGaussian = true;
      return (v1 * multiplier * sigma + mean);
    }
  }

}
