package StreamSimulator.StreamJoin;
/**
 * <p>Title: Li Feifei</p>
 * <p>Description: Stream Simulator for Stream Join Project</p>
 * <p>Copyright: Copyright (c) 2004 Computer Science Department, Boston University</p>
 * <p>Company: </p>
 * @author unascribed
 * @version 1.0
 */
import StreamSimulator.StreamGenerator.*;
import StreamSimulator.StreamGenerator.TimeCorrelation.*;
import StreamSimulator.Stream.*;
import StreamSimulator.Buffer.*;
import StreamSimulator.Clock.*;
import java.util.*;
import java.io.*;

public class StreamJoin {

  // Aging increment/decrement credits methods 
  public final static int PLUS_AND_MINUS_ONE = 1;
  public final static int PLUS_BY_OFFSET = 2;
  public final static int CONSTANT_TOTAL_CREDIT = 3;

  public static boolean SAMPLE = false;
  public static double SAMPLE_RATE = 0.7;

  static int BUFFER_TYPE = Topology.RANDOM_SHEDDING;
  public static  int STREAM_TYPE = Generator.PARETO;
  public static  int STREAM_TYPE1 = Generator.PARETO;
  public static  int STREAM_TYPE2 = Generator.PARETO;
  public static  int STREAM_TYPE3 = Generator.REVERSEPARETO;

  public static int NUM_ATTRIBUTE = 1;
  static int NUM_STREAMS = 4;  

  public static double AGING_INITIAL_VALUE = 0.998;
  public static int AGING_METHODS = 1;
  
  //public static int TIME_RANGE = 4320; //in time unit for weather
  //public static int WINDOW_SIZE = 24; //in time unit for weather
  //public static int[] BUFFERS={720, 720, 0, 0, 0, 0}; //for weather
  //public static int UNIT = 30;  //for weather
  
  //public static int TIME_RANGE = 2468; //in time unit for gismo
  //public static int WINDOW_SIZE = 24; //in time unit for gismo
  //public static int[] BUFFERS={168, 168, 0, 0, 0, 0}; //for gismo
  //public static int UNIT = 30; //for gismo
  
  //public static int TIME_RANGE = 36000; //in time unit for stock - ten hours
  //public static int WINDOW_SIZE = 60; //in time unit for stock - 1 minutes
  //public static int[] BUFFERS={480, 640, 1200, 1200, 1200, 0}; //for stock
  //public static int UNIT = 1000; //for stock - one second
  
  
  public static int TIME_RANGE = 10000; //in time unit for stock - ten hours
  public static int WINDOW_SIZE = 100; //in time unit for stock - 1 minutes
  public static int[] BUFFERS={80, 20}; //for stock
  public static int UNIT = 1; //for stock - one second


  static int DOMAIN_SIZE = 100;
  static int DOMAIN_MIN = 1;
  static int DOMAIN_SIZE1 = 100;
  static int DOMAIN_MIN1 = 1;
  static int BUFFER_SIZE = 50;
  static double SIGMA1 = 10;
  static double SIGMA2 = 10;
  static double SKEWNESS = 1.0;
  static double SKEWNESS1 = 1.0;  
  static double SKEWNESS2 = 1.0;
  static double SKEWNESS3 = 1.0;

  static long SEED_STEP = 4321123456789L;
  static boolean SAME_STREAM = true;
  static boolean SAME_TIME_STREAM = true;
  static boolean USE_HISTOGRAM = false;
  static long REBUILD_INTERVAL = 100;
  static long WARM_UP = 0;
  static int GRID_SIZE = 60;
  static int NUM_BUCKETS = 600;
  private long seed;
  static int SAME_ATTR =1;
  static int TIME_GENTYPE = Generator.UNIFORM;
  static int TIME_MEANSTEP = 3;
  static int TIME_SHIFT = 0;
  
  static boolean SEPERATE_BUFFER = false;
  
  static boolean SAVE_STREAMTOFILE = true;
  // number of tuples in a period used in stanford algorithm
  public static int STANFORD_PERIOD = 1;
  public static double STANFORD_THRESHOLD = 1.1;

  
  public StreamJoin() {
  }

  public long[][] simulate()
  {
      GregorianCalendar calendar = new GregorianCalendar();
      Date date = calendar.getTime();
      seed = date.getTime();
      Clock clock = new Clock(UNIT);
      Topology t = new Topology(this.BUFFER_TYPE,this.BUFFER_SIZE,this.WINDOW_SIZE,
         this.NUM_STREAMS, this.WARM_UP,this.USE_HISTOGRAM, this.SEPERATE_BUFFER,this.BUFFERS,STREAM_TYPE,TIME_RANGE);
      if (this.USE_HISTOGRAM)
        t.SetHistogram(this.REBUILD_INTERVAL,this.GRID_SIZE,this.NUM_BUCKETS);

      StreamSource[] stream=new StreamSource[this.NUM_STREAMS];
      int range = this.WINDOW_SIZE;
      
      
      /*for (int i =0; i<this.NUM_STREAMS;i++)
      {
        int arrivalRate = 1;
        int totalData = this.TIME_RANGE * arrivalRate;
        int shift = TIME_SHIFT;
        stream[i] = new StreamSource(STREAM_TYPE,DOMAIN_SIZE,DOMAIN_MIN,
        SKEWNESS, SIGMA1, SIGMA2, seed+i*SEED_STEP,arrivalRate,"data"+i,SAVE_STREAMTOFILE,i,
        SAME_ATTR, TIME_GENTYPE,totalData,TIME_MEANSTEP,shift, range);
        stream[i].set_topology(t);
        clock.addListener(stream[i]);
        t.streams.add(stream[i]);
        if (SAME_STREAM!=true)
          STREAM_TYPE = (STREAM_TYPE+1)%3;
        if (SAME_TIME_STREAM!=true)
          TIME_GENTYPE = (TIME_GENTYPE+1)%3;
        //shift += (int)(this.DOMAIN_SIZE*0.2);
      }*/
       int arrivalRate = 1;
        int totalData = this.TIME_RANGE * arrivalRate;
        int shift = TIME_SHIFT;
        stream[0] = new StreamSource(STREAM_TYPE,DOMAIN_SIZE,DOMAIN_MIN,
        SKEWNESS, SIGMA1, SIGMA2, seed+0*SEED_STEP,arrivalRate,"data"+0,SAVE_STREAMTOFILE,0,
        SAME_ATTR, TIME_GENTYPE,totalData,TIME_MEANSTEP,shift, range);
        stream[0].set_topology(t);
        clock.addListener(stream[0]);
        t.streams.add(stream[0]);

        arrivalRate = 1;
        totalData = this.TIME_RANGE * arrivalRate;
        shift = TIME_SHIFT;
        stream[1] = new StreamSource(STREAM_TYPE1,DOMAIN_SIZE1,DOMAIN_MIN1,
        SKEWNESS1, SIGMA1, SIGMA2, seed+1*SEED_STEP,arrivalRate,"data"+1,SAVE_STREAMTOFILE,1,
        SAME_ATTR, TIME_GENTYPE,totalData,TIME_MEANSTEP,shift, range);
        stream[1].set_topology(t);
        clock.addListener(stream[1]);
        t.streams.add(stream[1]);
        
        arrivalRate = 1;
        totalData = this.TIME_RANGE * arrivalRate;
        shift = TIME_SHIFT;
        stream[2] = new StreamSource(STREAM_TYPE2,DOMAIN_SIZE,DOMAIN_MIN,
        SKEWNESS2, SIGMA1, SIGMA2, seed+0*SEED_STEP,arrivalRate,"data"+2,SAVE_STREAMTOFILE,2,
        SAME_ATTR, TIME_GENTYPE,totalData,TIME_MEANSTEP,shift, range);
        stream[2].set_topology(t);
        clock.addListener(stream[2]);
        t.streams.add(stream[2]);

        arrivalRate = 1;
        totalData = this.TIME_RANGE * arrivalRate;
        shift = TIME_SHIFT;
        stream[3] = new StreamSource(STREAM_TYPE3,DOMAIN_SIZE1,DOMAIN_MIN1,
        SKEWNESS3, SIGMA1, SIGMA2, seed+1*SEED_STEP,arrivalRate,"data"+3,SAVE_STREAMTOFILE,3,
        SAME_ATTR, TIME_GENTYPE,totalData,TIME_MEANSTEP,shift, range);
        stream[3].set_topology(t);
        clock.addListener(stream[3]);
        t.streams.add(stream[3]);


      
      clock.setFinalListener(t);
      for (int i=0;i<TIME_RANGE;i++)
      {
        clock.nextTime();
      }
      for (int i=0;i<this.NUM_STREAMS;i++)
      {
        if (stream[i].gen.write_toFile == true)
        {
          try {
            stream[i].gen.outputFile.close();
          }
          catch (Exception e) {
            e.printStackTrace();
          }
        }
      }
      return t.joinResults;
  }


  public void getDataDistribtion(Generator gen)
  {
    Iterator it = gen.statistics.keySet().iterator();
    int cc=0;
    while(it.hasNext())
    {
      Integer key = (Integer)it.next();
      Integer freq = (Integer)gen.statistics.get(key);
      System.out.print(key.intValue() +"   " + freq.intValue()+"\n");
      cc+=freq.intValue();
    }
    System.out.println(gen.statistics.size()+" "+cc);
  }

  public static double combinationGen(int i, double s, int k, int n, double[] group)
  {
    int t;
    double tmpS=0.0;
    double sum=0.0;
    if (k==n)
      return s;
    k++;
    for (t=i+1;t<group.length;t++)
    {
      tmpS=s*group[t];
      sum+=combinationGen(t,tmpS,k,n,group);
    }
    return sum;
  }


  final static String usage = "\n -timeRange x | -bufferAlgo x | -streamType1 x "+
   "| -streamType2 x | -skewness x | -bufferSize x | -bufferSize1 x | -bufferSize2 x | -windowSize x | -domainSize x "+
   "| -domainMinVal x | -timedimension | -dependentType x | -shift x | -group x | -unit x | -meanStep x "+
   "| -sigma x | -dobufferAllocation | -agingInitialValue | -agingMethods\n";
  public static void main(String[] argv)
  {
    StreamJoin sim = new StreamJoin();
    for (int i = 0; i < argv.length; i++) {
      if (argv[i].equals("-timeRange")) {
          TIME_RANGE = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-streamType")) {
              STREAM_TYPE = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferAlgo")) {
              BUFFER_TYPE = (new Integer(argv[++i])).intValue();
      } else
        if (argv[i].equals("-skewness")) {
              SKEWNESS = (new Double(argv[++i])).doubleValue();
      } else
          if (argv[i].equals("-bufferSize")) {
              BUFFER_SIZE = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferSize1")) {
              BUFFERS[0] = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferSize2")) {
              BUFFERS[1] = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferSize3")) {
              BUFFERS[2] = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferSize4")) {
              BUFFERS[3] = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferSize5")) {
              BUFFERS[4] = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-bufferSize6")) {
              BUFFERS[5] = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-unit")) {
              UNIT = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-windowSize")) {
              WINDOW_SIZE = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-domainSize")) {
              DOMAIN_SIZE = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-domainMinVal")) {
              DOMAIN_MIN = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-shift")) {
              TIME_SHIFT = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-meanStep")) {
              TIME_MEANSTEP = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-agingMethods")) {
              AGING_METHODS = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-stanfordPeriod")) {
              STANFORD_PERIOD = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-numStreams")) {
              NUM_STREAMS = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-numAttribute")) {
              NUM_ATTRIBUTE = (new Integer(argv[++i])).intValue();
      } else
          if (argv[i].equals("-sigma1")) {
              SIGMA1 = (new Double(argv[++i])).doubleValue();
      } else
          if (argv[i].equals("-agingInitialValue")) {
              AGING_INITIAL_VALUE = (new Double(argv[++i])).doubleValue();
      } else
          if (argv[i].equals("-stanfordThreshold")) {
              STANFORD_THRESHOLD = (new Double(argv[++i])).doubleValue();
      } else
          if (argv[i].equals("-sampleRate")) {
          	  SAMPLE = true;
              SAMPLE_RATE = (new Double(argv[++i])).doubleValue();
      }
      else {
               System.out.println("Unknown parameter : " + argv[i]);
               System.out.println("USAGE : " + usage);
               System.exit(0);
      }
    }
      for (int i=0;i<1;i++)
      {
        try
        {
          long tmpRes[][] = new long[NUM_STREAMS][StreamJoin.NUM_ATTRIBUTE];
          for (int k=0;k<NUM_STREAMS;k++)
            for (int j=0;j<StreamJoin.NUM_ATTRIBUTE;j++)
              tmpRes[k][j]=0;
           tmpRes = sim.simulate();
           long total = 0;
           for (int k=0;k< NUM_STREAMS;k++)
           {
             for (int j=0;j<StreamJoin.NUM_ATTRIBUTE;j++)
             {  total+=tmpRes[k][j];
                System.out.print(tmpRes[k][j] + " ");
             }
             System.out.print("\n");
           }
           System.out.print(total+"\n");
        }
        catch(Exception e)
        {
          e.printStackTrace();
        }
      }
  }
}
